import inspect
from collections.abc import Callable, Iterable
from typing import Any, Protocol

from dependency_injector import containers, providers
from dependency_injector.wiring import Provide

type Container = type[containers.Container] | containers.Container

DEFAULT_RESOLVABLE_PROVIDER_TYPES = (providers.Factory, providers.Singleton)


def resolve_by_name_from_container_provider(
    name: str,
    container: providers.Container,
    provider_types: Iterable[type[providers.Provider]],
) -> providers.Provider | None:
    if not isinstance(name, str):
        raise ValueError("`name` must be string")

    for provider_name, provider in container.providers.items():  # type: ignore[union-attr]
        if not isinstance(provider, (providers.Container, *provider_types)):
            continue

        if provider_name == name:
            return provider

        if isinstance(provider, providers.Container):
            result = resolve_by_name_from_container_provider(
                name=name,
                container=provider,
                provider_types=provider_types,
            )
            if result is not None:
                return result

    return None


def resolve_by_name(
    name: str,
    container: Container,
    provider_types: Iterable[type[providers.Provider]],
) -> providers.Provider:
    if not isinstance(name, str):
        raise ValueError("`name` must be string")

    for provider_name, provider in container.providers.items():
        if not isinstance(provider, (providers.Container, *provider_types)):
            continue

        if provider_name == name:
            return provider

        if isinstance(provider, providers.Container):
            result = resolve_by_name_from_container_provider(
                name=name,
                container=provider,
                provider_types=provider_types,
            )
            if result is not None:
                return result

    raise ValueError(f"No provider named {name} in {container}")


def resolve_by_object_name(
    name: str,
    container: Container,
    provider_types: Iterable[type[providers.Provider]],
) -> Any:
    candidates: list[providers.Factory] = []
    for provider in container.traverse([*provider_types]):  # type: ignore[no-matching-overload]
        if not inspect.isclass(provider.cls):  # type: ignore[missing-attribute]
            continue

        if provider.cls.__name__.split(".")[-1] == name:  # type: ignore[missing-attribute]
            candidates.append(provider)  # type: ignore[bad-argument-type]

    if not candidates:
        raise ValueError(f"No object named {name} in {container}")

    if len(candidates) > 1:
        candidate_names = [p.cls.__name__ for p in candidates]  # type: ignore[union-attr]
        raise ValueError(f"Multiple objects named {name} in {container}: {candidate_names}")

    return candidates[0]()


def _collect_providers_by_type(
    type_: type,
    container: Container | providers.Container,
    provider_types: tuple[type[providers.Provider], ...],
) -> list[tuple[str, providers.Provider]]:
    candidates: list[tuple[str, providers.Provider]] = []
    for provider_name, provider in container.providers.items():  # type: ignore[union-attr]
        if isinstance(provider, providers.Container):
            candidates.extend(_collect_providers_by_type(type_, provider, provider_types))
        elif isinstance(provider, provider_types):
            provider_cls = provider.cls  # type: ignore[union-attr]
            if not (inspect.isclass(provider_cls) or inspect.ismethod(provider_cls)):
                continue
            cls = provider_cls.__self__ if inspect.ismethod(provider_cls) else provider_cls
            if cls.__name__ == type_.__name__:  # type: ignore[union-attr]
                candidates.append((provider_name, provider))
    return candidates


def resolve[T](
    type_: type[T],
    container: Container,
    provider_types: Iterable[type[providers.Provider]] = DEFAULT_RESOLVABLE_PROVIDER_TYPES,
    *,
    name: str | None = None,
) -> providers.Provider[T]:
    """
    의존성 전체 등록 경로를 참조하지 않고 의존성을 해결합니다. 다른 패키지의 의존성을 참조하는 경우
    의존 대상 선언 경로에 깊게 의존하는 것을 방지합니다. `container_cls`는 최상위 컨테이너일수도,
    의존성이 등록된 (하위) 컨테이너일수도 있습니다. 클래스 대상으로만 작동합니다.
    """
    if not inspect.isclass(type_):
        raise ValueError("Only class can be resolved")

    candidates = _collect_providers_by_type(type_, container, tuple(provider_types))

    if not candidates:
        raise ValueError(f"No provider found for {type_.__name__} in {container}")

    if len(candidates) == 1:
        return candidates[0][1]  # type: ignore[return-value]

    if name is None:
        candidate_names = [n for n, _ in candidates]
        raise ValueError(
            f"Multiple providers found for {type_.__name__}: {candidate_names}. "
            f"Specify `name` to select one"
        )

    for candidate_name, provider in candidates:
        if candidate_name == name:
            return provider  # type: ignore[return-value]

    candidate_names = [n for n, _ in candidates]
    raise ValueError(f"No provider named {name} for {type_.__name__}. Available: {candidate_names}")


def provide[T](
    type_: type[T],
    from_: type[containers.Container],
    /,
    *,
    provider_types: Iterable[type[providers.Provider]] = DEFAULT_RESOLVABLE_PROVIDER_TYPES,
    name: str | None = None,
) -> Provide[T]:
    return Provide[
        resolve(
            type_=type_,
            container=from_,
            provider_types=provider_types,
            name=name,
        )
    ]


def create_event_handler_resolver(container: Container) -> Callable[[type], Any]:
    async def resolver(handler_type: type) -> Any:
        return resolve_by_object_name(
            name=handler_type.__name__,
            container=container,
            provider_types=DEFAULT_RESOLVABLE_PROVIDER_TYPES,
        )

    return resolver


def create_resolver(container: Container) -> Callable[[str], Any]:
    def resolver(name: str) -> Any:
        return resolve_by_name(
            name=name, container=container, provider_types=DEFAULT_RESOLVABLE_PROVIDER_TYPES
        )

    return resolver


class Provider[T](Protocol):
    def __call__(
        self,
        t: type[T],
        /,
        name: str | None = None,
    ) -> Provide[T]: ...


def create_provider[T](
    container: type[containers.Container],
    provider_types: Iterable[type[providers.Provider]] = DEFAULT_RESOLVABLE_PROVIDER_TYPES,
) -> Provider[T]:
    def provider(
        type_: type[T],
        /,
        name: str | None = None,
    ) -> Provide[T]:
        return provide(
            type_,
            container,
            provider_types=provider_types,
            name=name,
        )

    return provider


def get_wiring_packages(*container_types: *tuple[type[containers.Container], ...]) -> set[str]:
    return {get_container_package(c) for c in container_types}


def get_container_package(container_type: type[containers.Container]) -> str:
    parts = container_type.__module__.split(".")

    if len(parts) >= 2:
        root, package, *_ = parts
        return f"{root}.{package}"

    raise ValueError(
        f"Invalid module path for {container_type.__name__}: expected 'root.package.*', got '{container_type.__module__}'"
    )



import http
import secrets
from collections.abc import Callable
from typing import Literal, Self

from fastapi import HTTPException, Request
from fastapi.security import APIKeyHeader as FastAPIAPIKeyHeader
from fastapi.security import HTTPBasic as FastAPIHTTPBasic
from fastapi.security import HTTPBearer as FastAPIHTTPBearer
from fastapi.security.base import SecurityBase
from starlette.exceptions import HTTPException as StarletteHTTPException


def unauthorized_error(detail: str, headers: dict[str, str] | None = None) -> HTTPException:
    return HTTPException(
        status_code=http.HTTPStatus.UNAUTHORIZED,
        detail=detail,
        headers=headers,
    )


class HTTPBasic[T](FastAPIHTTPBasic):
    def __init__(
        self,
        *,
        authenticator: Callable[[str, str], T],
        scheme_name: str | None = None,
        realm: str | None = None,
        description: str | None = None,
        auto_error: bool = True,
    ):
        super().__init__(
            scheme_name=scheme_name,
            realm=realm,
            description=description,
            auto_error=auto_error,
        )
        self.authenticator = authenticator

    async def __call__(self, request: Request) -> T:
        credentials = await super().__call__(request)
        return self.authenticator(credentials.username, credentials.password)

    @classmethod
    def static(cls, username: str, password: str) -> Self:
        def compare_digest(a: str, b: str) -> bool:
            return secrets.compare_digest(a.encode(), b.encode())

        def authenticate(_username: str, _password: str) -> None:
            is_correct_username = compare_digest(_username, username)
            is_correct_password = compare_digest(_password, password)

            if not (is_correct_username and is_correct_password):
                raise unauthorized_error(
                    detail="Incorrect username or password",
                    headers={"WWW-Authenticate": "Basic"},
                )

        return cls(authenticator=authenticate)


class APIKeyHeader(FastAPIAPIKeyHeader):
    def __init__(
        self,
        *,
        api_key: str,
        name: str = "X-API-Key",
        auto_error: bool = True,
        scheme_name: str | None = None,
        description: str | None = None,
    ):
        super().__init__(
            name=name,
            auto_error=auto_error,
            scheme_name=scheme_name,
            description=description,
        )
        self.api_key = api_key

    async def __call__(self, request: Request) -> str:
        exc = unauthorized_error(detail="Not authenticated. Invalid API key")

        try:
            input_api_key = await super().__call__(request)
        except (StarletteHTTPException, HTTPException):
            raise exc

        if input_api_key is None or input_api_key != self.api_key:
            raise exc

        return input_api_key


class HTTPBearer(FastAPIHTTPBearer):
    def __init__(
        self,
        *,
        auto_error: bool = True,
        error_handlers: dict[Literal["on_credentials_missing"], Callable[[], HTTPException]]
        | None = None,
    ):
        super().__init__(auto_error=auto_error)
        self.error_handlers = error_handlers or {}
        self._default_error_handler = lambda: unauthorized_error(detail="Not authenticated")

    async def __call__(self, request: Request) -> str:
        try:
            credentials = await super().__call__(request)
        except HTTPException:
            raise self.error_handlers.get("on_credentials_missing", self._default_error_handler)()

        return credentials.credentials


class CookieAuth(SecurityBase):
    def __init__(
        self,
        *,
        name: str,
        error_handlers: dict[Literal["on_cookie_missing"], Callable[[], HTTPException]]
        | None = None,
    ):
        self.name = name
        self.error_handlers = error_handlers or {}
        self._default_error_handler = lambda: unauthorized_error(detail="Not authenticated")

    async def __call__(self, request: Request) -> str:
        if (credentials := request.cookies.get(self.name)) is None:
            raise self.error_handlers.get("on_cookie_missing", self._default_error_handler)()

        return credentials



from .accumulator import *  # noqa: F401, F403
from .array import *  # noqa: F401, F403
from .array_expression import *  # noqa: F401, F403
from .comparison import *  # noqa: F401, F403
from .conditional import *  # noqa: F401, F403
from .set import *  # noqa: F401, F403
from .set_expression import *  # noqa: F401, F403
from .string import *  # noqa: F401, F403
from .string_expression import *  # noqa: F401, F403
from .type_expression import *  # noqa: F401, F403



from .exceptions import LockAcquisitionFailedError

__all__ = ["LockAcquisitionFailedError"]

try:
    from .mongo import MongoLock, MongoLockAcquirer

    __all__ += ["MongoLock", "MongoLockAcquirer"]
except ImportError:
    pass



import datetime
from dataclasses import dataclass
from typing import Any, Self

import pendulum
from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema


class DateTime(pendulum.DateTime):
    """https://stackoverflow.com/a/76719893"""

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(
            cls._validate, handler(datetime.datetime)
        )

    @classmethod
    def _validate(cls, v: datetime.datetime) -> pendulum.DateTime:
        return pendulum.instance(obj=v, tz=pendulum.UTC)


@dataclass
class DateRange:
    start: pendulum.DateTime
    end: pendulum.DateTime

    @classmethod
    def day_of(
        cls,
        dt: pendulum.DateTime,
        tz: pendulum.tz.Timezone | str = pendulum.UTC,
    ) -> Self:
        if dt.tzinfo is None:
            raise ValueError("input datetime must be aware")

        return cls(start=(start := dt.start_of("day").in_tz(tz)), end=start.add(days=1))



import asyncio
import contextlib
import logging
import signal
import threading
import time
import types
from collections.abc import Callable, Generator

from ..types import Seconds
from .consumer import MessageConsumerType
from .signal_handler import SignalHandler, ThreadSignalHandler

logger = logging.getLogger("compose")


class ThreadMessageConsumerRunner:
    def __init__(
        self,
        message_consumer_factory: Callable[[], MessageConsumerType],
        signal_handler_factory: Callable[[], SignalHandler] = ThreadSignalHandler,
    ):
        self.message_consumer_factory = message_consumer_factory
        self.signal_handler_factory = signal_handler_factory

        self._received_signal = False
        self._consumers: list[MessageConsumerType] = []

        for signum in (signal.SIGINT, signal.SIGTERM):
            signal.signal(signum, self.handle_signal)

    def run(self, num_workers: int = 1) -> None:
        threads = []

        for _ in range(num_workers):
            t = threading.Thread(target=self._run_in_thread)
            t.start()
            threads.append(t)

        while not self._received_signal:
            time.sleep(0.5)

        for consumer in self._consumers:
            consumer.shutdown()

        for t in threads:
            t.join()

    def _run_in_thread(self) -> None:
        with asyncio.Runner() as runner:
            runner.run(self._run_consumer())

    async def _run_consumer(self) -> None:
        message_consumer = self.message_consumer_factory()
        self._consumers.append(message_consumer)
        await message_consumer.run()

    def handle_signal(self, signum: int, _: types.FrameType) -> None:
        logger.info(f"Received {signal.Signals(signum).name}, exiting gracefully")
        self._received_signal = True


class FastAPIMessageConsumerRunner:
    def __init__(
        self,
        message_consumer_factory: Callable[[], MessageConsumerType],
        shutdown_timeout: Seconds = Seconds(30),
    ):
        self.message_consumer_factory = message_consumer_factory
        self.shutdown_timeout = shutdown_timeout

        self._consumer: MessageConsumerType | None = None
        self._thread: threading.Thread | None = None

    def _run_in_thread(self) -> None:
        try:
            self._consumer = self.message_consumer_factory()
            asyncio.run(self._consumer.run())
        except Exception as exc:
            logger.exception(f"Failed to run message consumer: {exc}")
            raise

    def run(self) -> None:
        self._thread = threading.Thread(target=self._run_in_thread)
        self._thread.start()

    def shutdown(self) -> None:
        if self._consumer is not None:
            self._consumer.shutdown()
        if self._thread is not None:
            self._thread.join(timeout=self.shutdown_timeout)

    @contextlib.contextmanager
    def lifespan(self) -> Generator[None, None, None]:
        self.run()
        yield
        self.shutdown()



import inspect
from collections.abc import Callable
from typing import Annotated, Any, Protocol

from dependency_injector.wiring import inject
from fastapi import Depends

from compose.di.dependency_injector.wiring import Provider


class HasSignature(Protocol):
    __signature__: inspect.Signature

    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...


def auto_wired[F: HasSignature](provider: Provider) -> Callable[[F], F]:
    """
    FastAPI 엔드포인트에 의존성을 자동으로 주입하는 데코레이터.

    주의: `__signature__` 속성은 일반 함수에만 존재하므로 해당 데코레이터는 메서드에 사용할 수 없습니다.
    """

    def wrapper(f: F) -> F:
        signature = inspect.signature(f)

        updated_params = []
        for name, param in signature.parameters.items():
            updated_param = param
            try:
                provided = provider(param.annotation)
                updated_param = updated_param.replace(
                    annotation=Annotated[param.annotation, Depends(provided)]
                )
            except ValueError:
                pass

            updated_params.append(updated_param)

        f.__signature__ = signature.replace(parameters=updated_params)
        f = inject(f)

        return f

    return wrapper



from __future__ import annotations

from typing import TYPE_CHECKING

import botocore.exceptions

if TYPE_CHECKING:
    import mypy_boto3_s3


def s3_obj_exists(s3_client: mypy_boto3_s3.S3Client, bucket: str, key: str) -> bool:
    try:
        s3_client.head_object(Bucket=bucket, Key=key)
        return True
    except botocore.exceptions.ClientError as exc:
        if exc.response["Error"]["Code"] == "404":
            return False

        raise exc



import functools
from collections.abc import Callable
from typing import Any

from pymongo.client_session import ClientSession, SessionOptions, TransactionOptions

from ..utils import unordered_partial


class MongoUnitOfWork[T]:
    def __init__(self, session_factory: Callable[..., ClientSession]):
        self.session_factory = session_factory

    def with_transaction(
        self,
        callback: Callable[[ClientSession], T] | Callable[..., T],
        *,
        session_options: SessionOptions | None = None,
        transaction_options: TransactionOptions | None = None,
        **kwargs: Any,
    ) -> T:
        session_options = session_options or SessionOptions()
        transaction_options = transaction_options or TransactionOptions()

        with self.session_factory(
            causal_consistency=session_options.causal_consistency,
            default_transaction_options=session_options.default_transaction_options,
            snapshot=session_options.snapshot,
        ) as session:
            result = session.with_transaction(
                callback=(
                    unordered_partial(p=functools.partial(callback), t=ClientSession)
                    if isinstance(callback, functools.partial)
                    else unordered_partial(p=functools.partial(callback, **kwargs), t=ClientSession)
                ),
                read_concern=transaction_options.read_concern,
                write_concern=transaction_options.write_concern,
                read_preference=transaction_options.read_preference,
                max_commit_time_ms=transaction_options.max_commit_time_ms,
            )

        return result


def mongo_transactional[T](func: Callable[..., T]) -> Callable[..., T]:
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> T:
        instance = args[0]

        if not hasattr(instance, "__dict__"):
            raise ValueError(f"`{instance.__class__.__name__}` does not have `__dict__` attribute")

        uow: MongoUnitOfWork | None = next(
            (t for t in instance.__dict__.values() if isinstance(t, MongoUnitOfWork)),
            None,
        )
        if uow is None:
            raise ValueError(
                f"`{instance.__class__.__name__}` does not have `{MongoUnitOfWork.__name__}` attribute"
            )

        return uow.with_transaction(functools.partial(func, *args, **kwargs))

    return wrapper



from .query import (
    MongoCursorPaginationQuery,
    MongoOffsetPaginationQuery,
    MongoPaginationQuery,
    MongoQuery,
)

__all__ = [
    "MongoCursorPaginationQuery",
    "MongoOffsetPaginationQuery",
    "MongoPaginationQuery",
    "MongoQuery",
]



try:
    from .instrumentation.loguru.instrumentor import LoguruInstrumentor
    from .tracer_provider import get_default_tracer_provider
except ImportError:
    raise ImportError("Install `opentelemetry` extra to use otel features") from None

__all__ = ["LoguruInstrumentor", "get_default_tracer_provider"]



import socket
import time
from typing import ClassVar, Self

from testcontainers.mongodb import MongoDbContainer as _MongoDbContainer


class MongoDbContainer(_MongoDbContainer):
    replica_set_name: ClassVar[str] = "rs"
    key_file_path: ClassVar[str] = "/tmp/mongodb-keyfile"

    def __init__(
        self,
        image: str = "mongo:latest",
        port: int = 27017,
        username: str | None = None,
        password: str | None = None,
        dbname: str | None = None,
        **kwargs,
    ):
        super().__init__(
            image=image,
            port=port,
            username=username,
            password=password,
            dbname=dbname,
            **kwargs,
        )
        self._with_replica_set = False

    def get_connection_url(self) -> str:
        connection_url = super().get_connection_url()
        return (
            f"{connection_url}/?replicaSet={self.replica_set_name}"
            if self._with_replica_set
            else connection_url
        )

    def with_replica_set(self) -> Self:
        self.port = self._get_available_port()
        self.with_bind_ports(container=self.port, host=self.port)
        self._command = [
            "bash",
            "-c",
            f"openssl rand -base64 756 > {self.key_file_path} && "
            f"chmod 400 {self.key_file_path} && "
            f"chown mongodb:mongodb {self.key_file_path} && "
            f"exec docker-entrypoint.sh mongod"
            f" --keyFile {self.key_file_path}"
            f" --replSet {self.replica_set_name}"
            f" --port {self.port}",
        ]
        self._with_replica_set = True

        return self

    def _connect(self) -> None:
        super()._connect()
        if self._with_replica_set:
            self._initialize_replica_set()

    def _initialize_replica_set(self) -> None:
        elapsed = 0
        timeout = 30
        interval = 1
        while True:
            exit_code, stdout = self.exec(
                f"mongosh "
                f"--username={self.username} "
                f"--password={self.password} "
                f"--eval=\"rs.initiate({{_id:'{self.replica_set_name}',members:[{{_id:0,host:'localhost:{self.port}'}}]}})\" "
                f"mongodb://localhost:{self.port}"
            )
            if not exit_code:
                break

            time.sleep(interval)
            elapsed += interval
            if elapsed > timeout:
                raise RuntimeError(f"Could not initiate replica set due to {stdout.decode()}")

    @staticmethod
    def _get_available_port() -> int:
        with socket.socket() as sock:
            sock.bind(("", 0))
            return sock.getsockname()[1]



from __future__ import annotations

from collections.abc import Callable
from typing import Any, Self

import pendulum

from compose import types

from . import utils
from .base import ComparisonOperator, EqualityOperator, Operator
from .types import DictExpression


class SkipNull(Operator):
    def __init__(self, op: ComparisonOperator):
        self.op = op

    def expression(self) -> dict[str, Any]:
        return self.op.expression() if self.op.value is not None else {}


def _expression_factory(mongo_operator: str) -> Callable[[ComparisonOperator], DictExpression]:
    def expression(self: ComparisonOperator) -> DictExpression:
        return {self.field: {mongo_operator: self.value}}

    return expression


def create_comparison_operator[T](
    name: str, mongo_operator: str, base: tuple[type[T], ...]
) -> type[T]:
    return utils.create_operator(
        name=name,
        base=base,
        expression_factory=_expression_factory(mongo_operator),
    )


Eq = create_comparison_operator(name="Eq", mongo_operator="$eq", base=(EqualityOperator,))
Ne = create_comparison_operator(name="Ne", mongo_operator="$ne", base=(EqualityOperator,))
Gt = create_comparison_operator(name="Gt", mongo_operator="$gt", base=(ComparisonOperator,))
Gte = create_comparison_operator(name="Gte", mongo_operator="$gte", base=(ComparisonOperator,))
Lt = create_comparison_operator(name="Lt", mongo_operator="$lt", base=(ComparisonOperator,))
Lte = create_comparison_operator(name="Lte", mongo_operator="$lte", base=(ComparisonOperator,))
In = create_comparison_operator(name="In", mongo_operator="$in", base=(ComparisonOperator,))
Nin = create_comparison_operator(name="Nin", mongo_operator="$nin", base=(ComparisonOperator,))


class Regex(ComparisonOperator):
    def __init__(
        self,
        field: str,
        value: Any | None = None,
        options: str = "ms",
    ):
        super().__init__(field=field, value=value)
        self.options = options

    def expression(self) -> dict[str, Any]:
        return {self.field: {"$regex": self.value, "$options": self.options}}


class Range(Operator):
    def __init__(self, g: Gt | Gte, l: Lt | Lte):  # noqa: E741
        self.g = g
        self.l = l

    def expression(self) -> dict[str, Any]:
        field = self.g.field
        return {field: self.g.expression()[field] | self.l.expression()[field]}

    @classmethod
    def date(cls, field: str, start: pendulum.DateTime, end: pendulum.DateTime) -> Self:
        return cls(g=Gte(field=field, value=start), l=Lt(field=field, value=end))

    @classmethod
    def day_of(cls, field: str, dt: pendulum.DateTime) -> Self:
        date_range = types.DateRange.day_of(dt)
        return cls.date(field=field, start=date_range.start, end=date_range.end)



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Self

from pydantic.fields import FieldInfo
from pydantic_settings import BaseSettings as PydanticBaseSettings
from pydantic_settings import PydanticBaseSettingsSource

from . import func
from .enums import AppEnv

if TYPE_CHECKING:
    import mypy_boto3_ssm


type SettingPreprocessor = Callable[[str], str]
type SettingPreprocessors = dict[str, SettingPreprocessor]


class ParameterStoreParameterPrefix(str):
    @classmethod
    def create(cls, service_name: str, app_env: AppEnv) -> Self:
        return cls(f"/{service_name}/{app_env.lower()}/")


def get_parameters_from_ssm(
    ssm_client: mypy_boto3_ssm.SSMClient,
    prefix: str,
    preprocessors: SettingPreprocessors | None = None,
) -> dict[str, Any]:
    preprocessors = preprocessors or {}

    result = {}
    has_next = True
    next_token: str | None = None
    while has_next:
        params = {
            "Path": prefix,
            "WithDecryption": True,
            "Recursive": True,
        }
        if next_token is not None:
            params["NextToken"] = next_token

        response = ssm_client.get_parameters_by_path(**params)
        for parameter in response["Parameters"]:
            key = parameter["Name"].split("/")[-1].lower()
            value = parameter["Value"]
            result[key] = preprocessors.get(key, func.ident)(value)

        next_token = response.get("NextToken")
        has_next = next_token is not None
    return result


class AWSParameterStoreSettingsSource(PydanticBaseSettingsSource):
    def __init__(
        self,
        settings_cls: type[PydanticBaseSettings],
        ssm_client: mypy_boto3_ssm.SSMClient,
        prefix: str,
        preprocessors: dict[str, Callable[[str], str]] | None = None,
    ):
        super().__init__(settings_cls=settings_cls)

        self.ssm_client = ssm_client
        self.prefix = prefix
        self.preprocessors = preprocessors or {}

        self._parameters = get_parameters_from_ssm(
            ssm_client=self.ssm_client,
            prefix=self.prefix,
            preprocessors=self.preprocessors,
        )

    def get_field_value(self, field: FieldInfo, field_name: str) -> tuple[Any, str, bool]:
        return self._parameters.get(field_name), field_name, False

    def __call__(self) -> dict[str, Any]:
        result = {}

        for field_name, field in self.settings_cls.model_fields.items():
            field_value, field_key, is_value_complex = self.get_field_value(
                field=field, field_name=field_name
            )
            field_value = self.prepare_field_value(
                field_name=field_name,
                field=field,
                value=field_value,
                value_is_complex=is_value_complex,
            )
            if field_value is None:
                continue

            result[field_name] = field_value

        return result

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(prefix={self.prefix})"


def create_base_settings(
    ssm_client: mypy_boto3_ssm.SSMClient,
    prefix: str,
    preprocessors: SettingPreprocessors | None = None,
) -> type[PydanticBaseSettings]:
    class BaseSettings(PydanticBaseSettings):
        @classmethod
        def settings_customise_sources(
            cls,
            settings_cls: type[PydanticBaseSettings],
            init_settings: PydanticBaseSettingsSource,
            env_settings: PydanticBaseSettingsSource,
            dotenv_settings: PydanticBaseSettingsSource,
            file_secret_settings: PydanticBaseSettingsSource,
        ) -> tuple[PydanticBaseSettingsSource, ...]:
            return (
                init_settings,
                AWSParameterStoreSettingsSource(
                    settings_cls=settings_cls,
                    ssm_client=ssm_client,
                    prefix=prefix,
                    preprocessors=preprocessors,
                ),
                env_settings,
                file_secret_settings,
            )

    return BaseSettings



from __future__ import annotations

from typing import IO, TYPE_CHECKING, Any

import botocore.exceptions

if TYPE_CHECKING:
    import mypy_boto3_s3


class S3ObjectStore:
    def __init__(self, s3_client: mypy_boto3_s3.S3Client, bucket: str) -> None:
        self._client = s3_client
        self._bucket = bucket

    def upload(self, key: str, body: bytes | IO[bytes], **params: Any) -> None:
        self._client.put_object(Bucket=self._bucket, Key=key, Body=body, **params)

    def download(self, key: str) -> bytes:
        response = self._client.get_object(Bucket=self._bucket, Key=key)
        return response["Body"].read()

    def delete(self, key: str) -> None:
        self._client.delete_object(Bucket=self._bucket, Key=key)

    def copy(self, src_key: str, dst_key: str) -> None:
        self._client.copy_object(
            Bucket=self._bucket,
            Key=dst_key,
            CopySource={"Bucket": self._bucket, "Key": src_key},
        )

    def exists(self, key: str) -> bool:
        try:
            self._client.head_object(Bucket=self._bucket, Key=key)
            return True
        except botocore.exceptions.ClientError as exc:
            if exc.response["Error"]["Code"] == "404":
                return False
            raise exc



from typing import Any

from pydantic import Field

from .. import model


class AuthorizationGrant(model.BaseModel):
    access_token: str
    refresh_token: str | None = None


class UserResource(model.BaseModel):
    email: str
    name: str


class TokenClaims(model.BaseModel):
    sub: str
    iss: str
    jti: str
    iat: int
    exp: int
    extra: dict[str, Any] = Field(default_factory=dict)



try:
    from .enums import *  # noqa: F403
    from .fixture import *  # noqa: F403
    from .hook import *  # noqa: F403
    from .param import *  # noqa: F403
except ImportError:
    raise ImportError("Install `pytest` to use testing fixtures") from None



from __future__ import annotations

from typing import Any, overload

from pymongo.client_session import ClientSession

from ..entity import Entity
from .mongo import MongoRepository


class _Finder:
    def __init__(self, field: str) -> None:
        self._field = field

    @overload
    def __get__[T: Entity](self, obj: None, objtype: type[MongoRepository[T]]) -> _Finder: ...

    @overload
    def __get__[T: Entity](
        self, obj: MongoRepository[T], objtype: type[MongoRepository[T]]
    ) -> _BoundFinder[T]: ...

    def __get__[T: Entity](
        self, obj: MongoRepository[T] | None, objtype: type[MongoRepository[T]]
    ) -> _Finder | _BoundFinder[T]:
        if obj is None:
            return self
        return _BoundFinder(obj, self._field)


class _BoundFinder[T: Entity]:
    def __init__(self, repo: MongoRepository[T], field: str) -> None:
        self._repo = repo
        self._field = field

    def __call__(self, value: Any, *, session: ClientSession | None = None) -> T | None:
        return self._repo.find_by({self._field: value}, session=session)


class _Lister:
    def __init__(self, field: str) -> None:
        self._field = field

    @overload
    def __get__[T: Entity](self, obj: None, objtype: type[MongoRepository[T]]) -> _Lister: ...

    @overload
    def __get__[T: Entity](
        self, obj: MongoRepository[T], objtype: type[MongoRepository[T]]
    ) -> _BoundLister[T]: ...

    def __get__[T: Entity](
        self, obj: MongoRepository[T] | None, objtype: type[MongoRepository[T]]
    ) -> _Lister | _BoundLister[T]:
        if obj is None:
            return self
        return _BoundLister(obj, self._field)


class _BoundLister[T: Entity]:
    def __init__(self, repo: MongoRepository[T], field: str) -> None:
        self._repo = repo
        self._field = field

    def __call__(
        self,
        value: Any,
        *,
        sort: list[tuple[str, int]] | None = None,
        session: ClientSession | None = None,
    ) -> list[T]:
        return self._repo.list_by({self._field: value}, sort=sort, session=session)


def finder(field: str) -> _Finder:
    return _Finder(field)


def lister(field: str) -> _Lister:
    return _Lister(field)



from __future__ import annotations

from typing import Any, Self

from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict

from . import field, types


class BaseModel(PydanticBaseModel):
    @classmethod
    def from_model(cls, model: BaseModel) -> Self:
        return cls.model_validate(model.model_dump())

    def validated_copy(
        self,
        *,
        update: dict[str, Any] | None = None,
        deep: bool = False,
    ) -> Self:
        result = self.model_copy(update=update, deep=deep)
        return self.__class__.model_validate(result.model_dump())

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        extra="ignore",
    )


class TimeStampedModel(BaseModel):
    created_at: types.DateTime = field.DateTimeField()
    updated_at: types.DateTime = field.DateTimeField()



from collections.abc import Callable
from typing import ClassVar, Self

from authlib.integrations.httpx_client import AsyncOAuth2Client

from . import vo


class AuthorizationServer:
    headers: ClassVar[dict[str, str]] = {"Content-Type": "application/x-www-form-urlencoded"}

    def __init__(
        self,
        auth_client_factory: Callable[..., AsyncOAuth2Client],
        access_token_url: str,
    ):
        self.auth_client_factory = auth_client_factory
        self.access_token_url = access_token_url

    @classmethod
    def default(cls, client_id: str, client_secret: str, access_token_url: str) -> Self:
        return cls(
            auth_client_factory=lambda: AsyncOAuth2Client(
                client_id=client_id,
                client_secret=client_secret,
            ),
            access_token_url=access_token_url,
        )

    async def grant_authorization(self, redirect_uri: str, code: str) -> vo.AuthorizationGrant:
        async with self.auth_client_factory() as client:
            response = await client.fetch_token(
                url=self.access_token_url,
                headers=self.headers,
                grant_type="authorization_code",
                client_id=client.client_id,
                client_secret=client.client_secret,
                redirect_uri=redirect_uri,
                code=code,
            )
        return vo.AuthorizationGrant.model_validate(response)

    async def renew_token(self, token: str) -> vo.AuthorizationGrant:
        async with self.auth_client_factory() as client:
            response = await client.fetch_token(
                url=self.access_token_url,
                headers=self.headers,
                grant_type="refresh_token",
                client_id=client.client_id,
                client_secret=client.client_secret,
                refresh_token=token,
            )
        return vo.AuthorizationGrant.model_validate(response)



from .extra import schema_by_field_name, schema_excludes
from .schema import (
    CursorListSchema,
    Error,
    Id,
    InvalidParam,
    ListSchema,
    Schema,
    TimeStampedSchema,
)

__all__ = [
    "CursorListSchema",
    "Error",
    "Id",
    "InvalidParam",
    "ListSchema",
    "Schema",
    "TimeStampedSchema",
    "schema_by_field_name",
    "schema_excludes",
]



try:
    from .endpoint import SpecialEndpoint, add_health_check_endpoint, health_check
    from .exception_handler import (
        ExceptionHandler,
        ExceptionHandlerInfo,
        create_exception_handler,
        default_exception_handlers,
    )
    from .openapi import (
        OpenAPIDoc,
        OpenAPISchema,
        RedocHTML,
        SwaggerUIHTML,
        add_doc_routes,
        additional_responses,
        openapi_tags,
    )
    from .otel import NonInstrumentedUrls
    from .param import FromAuth, FromPath, OffsetPaginationParams, as_query, with_fields
    from .response import NoContentResponse, ZipStreamingResponse
    from .routing import APIRouter, create_auto_wired_route
    from .security import APIKeyHeader, CookieAuth, HTTPBasic, HTTPBearer, unauthorized_error
    from .wiring import auto_wired
except ImportError:
    raise ImportError("Install `fastapi` to use `compose.fastapi` package") from None

__all__ = [
    "APIKeyHeader",
    "APIRouter",
    "CookieAuth",
    "ExceptionHandler",
    "ExceptionHandlerInfo",
    "FromAuth",
    "FromPath",
    "HTTPBasic",
    "HTTPBearer",
    "NoContentResponse",
    "NonInstrumentedUrls",
    "OffsetPaginationParams",
    "OpenAPIDoc",
    "OpenAPISchema",
    "RedocHTML",
    "SpecialEndpoint",
    "SwaggerUIHTML",
    "ZipStreamingResponse",
    "add_doc_routes",
    "add_health_check_endpoint",
    "additional_responses",
    "as_query",
    "auto_wired",
    "create_auto_wired_route",
    "create_exception_handler",
    "default_exception_handlers",
    "health_check",
    "openapi_tags",
    "unauthorized_error",
    "with_fields",
]


try:
    from .utils import (  # noqa: F401
        ErrorEvent,
        Level,
        capture_error,
        create_before_send_hook,
        init_sentry,
    )

    __all__.extend(
        [
            "ErrorEvent",
            "Level",
            "capture_error",
            "create_before_send_hook",
            "init_sentry",
        ]
    )
except ImportError:
    pass

try:
    from .dishka import injected_route  # noqa: F401

    __all__.append("injected_route")
except ImportError:
    pass



from .base import ListExpression, Operator


class Pipeline(Operator):
    def __init__(self, *ops: Operator):
        self.ops = list(ops)

    def expression(self) -> ListExpression:
        result: ListExpression = []
        for op in self.ops:
            expr = op.expression()
            if not expr:
                continue
            if isinstance(expr, list):
                result.extend(expr)
            else:
                result.append(expr)
        return result



from . import (
    asyncio,
    auth,
    command,
    concurrent,
    constants,
    dependency,
    di,
    entity,
    enums,
    event,
    exceptions,
    field,
    func,
    gunicorn,
    lock,
    messaging,
    monitoring,
    pagination,
    query,
    repository,
    schema,
    settings,
    stream,
    types,
    typing,
    uow,
    utils,
)
from .model import BaseModel, TimeStampedModel

tp = typing

__all__ = [
    "BaseModel",
    "TimeStampedModel",
    "asyncio",
    "auth",
    "command",
    "concurrent",
    "constants",
    "dependency",
    "di",
    "entity",
    "enums",
    "event",
    "exceptions",
    "field",
    "func",
    "gunicorn",
    "lock",
    "messaging",
    "monitoring",
    "pagination",
    "query",
    "repository",
    "schema",
    "settings",
    "stream",
    "tp",
    "types",
    "typing",
    "uow",
    "utils",
]

try:
    from . import logging  # noqa: F401

    __all__.append("logging")
except ImportError:
    pass

try:
    from . import testing  # noqa: F401

    __all__.append("testing")
except ImportError:
    pass

try:
    from . import fastapi  # noqa: F401

    __all__.append("fastapi")
except ImportError:
    pass

try:
    from . import otel  # noqa: F401

    __all__.append("otel")
except ImportError:
    pass

try:
    from . import httpx  # noqa: F401

    __all__.append("httpx")
except ImportError:
    pass

try:
    from . import aws  # noqa: F401

    __all__.append("aws")
except ImportError:
    pass

try:
    from . import testcontainers  # noqa: F401

    __all__.append("testcontainers")
except ImportError:
    pass



from collections.abc import Callable

from dependency_injector import providers


def factory_of_factory[**P, T](
    t: Callable[P, T],
    *args: P.args,
    **kwargs: P.kwargs,
) -> providers.Factory[providers.Factory[T]]:
    return providers.Factory(
        providers.Factory,
        t,
        *args,
        **kwargs,
    )



from typing import Any, Self

from .. import constants
from . import primitive


class Seconds(primitive.Float):
    def __new__(cls, v: Any, /) -> Self:
        v = super().__new__(cls, v)
        if v < 0:
            raise ValueError(f"`{cls.__name__}` must be a non-negative float")
        return v

    @classmethod
    def from_hours(cls, hours: float) -> Self:
        return cls(hours * constants.SECONDS_PER_HOUR)

    @classmethod
    def from_minutes(cls, minutes: float) -> Self:
        return cls(minutes * constants.SECONDS_PER_MINUTE)


class MilliSeconds(primitive.Int):
    def __new__(cls, v: Any, /) -> Self:
        v = super().__new__(cls, v)
        if v < 0:
            raise ValueError(f"`{cls.__name__}` must be a non-negative integer")
        return v

    @classmethod
    def from_seconds(cls, seconds: float) -> Self:
        return cls(seconds * constants.MILLISECONDS_PER_SECOND)



from __future__ import annotations

import functools
from typing import Any, ClassVar, Self, get_args, get_origin, overload

import pendulum
import pymongo
from pymongo.client_session import ClientSession
from pymongo.collection import Collection
from pymongo.database import Database
from pymongo.results import UpdateResult

from .. import types
from ..entity import Entity
from ..query.mongo import MongoPaginationQuery, MongoQuery
from ..utils import descendants_of
from .base import BaseRepository


class MongoDocument(dict[str, Any]):
    @classmethod
    def from_entity(cls, entity: Entity, **kwargs: Any) -> Self:
        return cls(
            entity.model_dump(
                by_alias=True,
                **{key: value for key, value in kwargs.items() if key not in {"by_alias"}},
            )
        )


class MongoRepository[T: Entity](BaseRepository):
    """
    `MongoRepository` 추상 클래스

    상속 인자:
        `session_requirement`: `MongoRepository` 상속 시 `session_requirement` 인자를 통해
        `session` 인자 선언 여부를 검사할 수 있습니다. 프로덕션에서는 `session_requirement`를
        `SessionRequirement.REQUIRED`로 설정하는 것을 권장합니다.

        ```python
        class UserRepository(
            MongoRepository[User],
            session_requirement=SessionRequirement.REQUIRED,
        ):
            ...
        ```

    """

    __collection_name__: ClassVar[str] = ""
    __indexes__: ClassVar[list[pymongo.IndexModel] | None] = None

    def __init__(self, collection: Collection):
        self.collection = collection

    @classmethod
    def create(cls, database: Database, **kwargs) -> Self:
        collection = database.get_collection(cls.__collection_name__, **kwargs)
        if (
            cls.__collection_name__ not in database.list_collection_names()
            and cls.__indexes__ is not None
        ):
            collection.create_indexes(cls.__indexes__)

        return cls(collection=collection)

    def find_by_id(
        self,
        entity_id: types.PyObjectId,
        session: ClientSession | None = None,
        **kwargs,
    ) -> T | None:
        return self.find_by({"_id": entity_id}, session=session, **kwargs)

    @overload
    def find_by(
        self,
        filter_: dict[str, Any],
        *,
        projection: None = None,
        session: ClientSession | None = None,
        **kwargs,
    ) -> T | None: ...

    @overload
    def find_by(
        self,
        filter_: dict[str, Any],
        *,
        projection: dict[str, Any],
        session: ClientSession | None = None,
        **kwargs,
    ) -> dict[str, Any] | None: ...

    def find_by(
        self,
        filter_: dict[str, Any],
        *,
        projection: dict[str, Any] | None = None,
        session: ClientSession | None = None,
        **kwargs,
    ) -> T | dict[str, Any] | None:
        validate_to_entity = projection is None
        query_result = self.collection.find_one(
            filter=filter_,
            session=session,
            **kwargs,
        )
        if query_result is None:
            return None

        return (
            self._entity_type.model_validate(query_result) if validate_to_entity else query_result
        )

    def find_by_query(
        self, qry: MongoQuery, session: ClientSession | None = None, **kwargs
    ) -> dict[str, Any] | None:
        query_result = self.collection.aggregate(qry.to_query(), session=session, **kwargs)
        return next(query_result, None)

    @overload
    def list_by(
        self,
        filter_: dict[str, Any],
        *,
        projection: None = None,
        sort: list[tuple[str, int]] | None = None,
        session: ClientSession | None = None,
        **kwargs,
    ) -> list[T]: ...

    @overload
    def list_by(
        self,
        filter_: dict[str, Any],
        *,
        projection: dict[str, Any],
        sort: list[tuple[str, int]] | None = None,
        session: ClientSession | None = None,
        **kwargs,
    ) -> list[dict[str, Any]]: ...

    def list_by(
        self,
        filter_: dict[str, Any],
        *,
        projection: dict[str, Any] | None = None,
        sort: list[tuple[str, int]] | None = None,
        session: ClientSession | None = None,
        **kwargs,
    ) -> list[T] | list[dict[str, Any]]:
        validate_to_entity = projection is None
        query_result = self.collection.find(
            filter=filter_,
            projection=projection,
            sort=sort,
            session=session,
            **kwargs,
        )
        return (
            [self._entity_type.model_validate(item) for item in query_result]
            if validate_to_entity
            else list(query_result)
        )

    def list_by_query(
        self, qry: MongoQuery, session: ClientSession | None = None, **kwargs
    ) -> list[dict[str, Any]]:
        query_result = self.collection.aggregate(qry.to_query(), session=session, **kwargs)
        return list(query_result)

    def add(self, entity: T, session: ClientSession | None = None, **kwargs) -> None:
        self.collection.insert_one(MongoDocument.from_entity(entity), session=session, **kwargs)

    def add_many(self, entities: list[T], session: ClientSession | None = None, **kwargs) -> None:
        self.collection.insert_many(
            [MongoDocument.from_entity(entity) for entity in entities], session=session, **kwargs
        )

    def update(self, entity: T, session: ClientSession | None = None, **kwargs) -> None:
        document = MongoDocument.from_entity(entity)
        update_result = self.collection.update_one(
            {"_id": entity.id},
            {"$set": document},
            session=session,
            **kwargs,
        )

        self.on_update(
            update_result=update_result,
            entity=entity,
            session=session,
            **kwargs,
        )

    def on_update(
        self,
        update_result: UpdateResult,
        entity: T,
        session: ClientSession | None = None,
        **kwargs,
    ) -> None:
        if not update_result.modified_count:
            return

        if getattr(entity, "updated_at") is None:
            return

        self.collection.update_one(
            {"_id": entity.id},
            {"$set": {"updated_at": pendulum.DateTime.utcnow()}},
            session=session,
            **kwargs,
        )

    def update_many(
        self, entities: list[T], session: ClientSession | None = None, **kwargs
    ) -> None:
        for entity in entities:
            self.update(entity=entity, session=session, **kwargs)

    def delete(
        self, entity_id: types.PyObjectId, session: ClientSession | None = None, **kwargs
    ) -> None:
        self.collection.delete_one({"_id": entity_id}, session=session, **kwargs)

    def paginate[R](
        self, qry: MongoPaginationQuery[R], session: ClientSession | None = None, **kwargs
    ) -> R:
        query_result = self.collection.aggregate(qry.to_query(), session=session, **kwargs)
        return qry.to_result(next(query_result, None))

    @functools.cached_property
    def _entity_type(self) -> T:
        orig_base = next(
            (base for base in self.__class__.__orig_bases__ if get_origin(base) is MongoRepository),  # type: ignore[missing-attribute]
            None,
        )
        if orig_base is None:
            raise ValueError(
                f"{self.__class__.__name__} must inherit MongoRepository[T] with explicit type parameter"
            )

        return get_args(orig_base)[0]


def setup_indexes(*databases: *tuple[Database, ...]) -> None:
    index_map: dict[str, list[pymongo.IndexModel]] = {}
    for subclass in descendants_of(MongoRepository):
        collection_name = subclass.__collection_name__
        indexes = subclass.__indexes__ or []
        if not (collection_name and indexes):
            continue
        collected = index_map.setdefault(collection_name, [])
        collected_documents = [idx.document for idx in collected]
        collected.extend(idx for idx in indexes if idx.document not in collected_documents)

    for database in databases:
        collection_names = database.list_collection_names()
        for collection_name, indexes in index_map.items():
            if collection_name not in collection_names:
                continue
            database.get_collection(collection_name).create_indexes(indexes)



from collections.abc import Sequence
from typing import Any, Protocol, Self, get_args

from pydantic import ConfigDict

from .. import model
from ..pagination import CursorPaginationResult, OffsetPaginationResult
from .extra import schema_by_field_name


class Schema(model.BaseModel):
    model_config = ConfigDict(json_schema_extra=schema_by_field_name())


class Id[T](Schema):
    id: T


class TimeStampedSchema(model.TimeStampedModel, Schema): ...


class ListSchema[T](Schema):
    total: int
    items: list[T]

    @classmethod
    def from_result(
        cls,
        result: OffsetPaginationResult,
        parser_name: str = "model_validate",
        **parser_kwargs: Any,
    ) -> Self:
        if not result.items:
            return cls(**result.model_dump())

        annotation = cls.model_fields["items"].annotation
        item_type = get_args(annotation)[0]

        if not issubclass(item_type, model.BaseModel):
            data = result.model_dump(exclude={"extra"}) | result.extra
            return cls(**data)

        if (parser := getattr(item_type, parser_name, None)) is None:
            raise AttributeError(f"{item_type.__name__} has no attribute: {parser_name}")

        return cls(
            **result.model_dump(exclude={"items", "extra"}),
            **result.extra,
            items=[parser(item, **parser_kwargs) for item in result.items],
        )

    @classmethod
    def from_items(cls, items: list[Any]) -> Self:
        return cls(total=len(items), items=items)


class CursorListSchema[T](Schema):
    items: list[T]
    next_cursor: str | None = None

    @classmethod
    def from_result(
        cls,
        result: CursorPaginationResult,
        parser_name: str = "model_validate",
        **parser_kwargs: Any,
    ) -> Self:
        if not result.items:
            return cls(items=[])

        annotation = cls.model_fields["items"].annotation
        item_type = get_args(annotation)[0]

        if not issubclass(item_type, model.BaseModel):
            return cls(**result.model_dump())

        if (parser := getattr(item_type, parser_name, None)) is None:
            raise AttributeError(f"{item_type.__name__} has no attribute: {parser_name}")

        return cls(
            items=[parser(item, **parser_kwargs) for item in result.items],
            next_cursor=result.next_cursor,
        )


class InvalidParam(model.BaseModel):
    loc: str
    message: str
    type: str


class HasErrors(Protocol):
    def errors(self) -> Sequence[Any]: ...


class Error(model.BaseModel):
    title: str
    type: str
    detail: str | None = None
    invalid_params: list[InvalidParam] | None = None

    @classmethod
    def from_validation_error(cls, exc: HasErrors, title: str) -> Self:
        invalid_params = []
        for error in exc.errors():
            invalid_params.append(
                InvalidParam(
                    loc=".".join(str(v) for v in error["loc"]),
                    message=error["msg"],
                    type=error["type"],
                )
            )
        return cls(
            title=title,
            type="validation_error",
            invalid_params=invalid_params,
        )



import base64
from typing import Any, Self

import pymongo

from compose.model import BaseModel

from .aggregation import Push
from .comparison import Eq, Gt, Lt
from .logical import And
from .pipeline import Pipeline
from .stage import EmptyStage, Group, Limit, Match, Project, Sort, Spec, Stage
from .types import DictExpression, ListExpression, _PositiveInt


class Cursor(BaseModel):
    def to_str(self) -> str:
        return base64.b64encode(self.model_dump_json(by_alias=True).encode()).decode()

    def to_dict(self) -> dict[str, Any]:
        return self.model_dump(by_alias=True)

    @classmethod
    def from_str(cls, cursor: str) -> Self:
        return cls.model_validate_json(json_data=base64.b64decode(cursor).decode())


class CursorPaginationClause(And):
    direction_to_op = {pymongo.ASCENDING: Gt, pymongo.DESCENDING: Lt}
    default_direction = pymongo.ASCENDING

    @classmethod
    def from_cursor_params(
        cls,
        sort_field: str,
        cursor_params: list[tuple[str, Any]],
        sort: Sort,
    ) -> Self:
        sort_params = {criterion.field: criterion.direction for criterion in sort.criteria}
        return cls(
            *(
                cls.direction_to_op[sort_params.get(field, cls.default_direction)](
                    field=field,
                    value=value,
                )
                if field == sort_field
                else Eq(field=field, value=value)
                for field, value in cursor_params
            )
        )


class CursorQuery(Stage[DictExpression]):
    def __init__(self, sort: Sort, cursor: Cursor | None = None):
        self.sort = sort
        self.cursor = cursor

    def expression(self) -> DictExpression:
        if (cursor := self.cursor) is None:
            return EmptyStage().expression()

        clauses = []
        cursor_params = cursor.to_dict().items()
        for idx, (field, _) in enumerate(cursor_params):
            current_cursor_params = list(cursor_params)[: idx + 1]
            clauses.append(
                CursorPaginationClause.from_cursor_params(
                    sort_field=field,
                    cursor_params=current_cursor_params,
                    sort=self.sort,
                )
            )

        return Match.or_(*clauses).expression()


class CursorPagination(Stage[ListExpression]):
    def __init__(
        self,
        sort: Sort,
        cursor: Cursor | None = None,
        per_page: int = 10,
    ):
        self.sort = sort
        self.cursor = cursor
        self.per_page = _PositiveInt(per_page)

    def expression(self) -> ListExpression:
        return Pipeline(
            CursorQuery(sort=self.sort, cursor=self.cursor),
            self.sort,
            Limit(self.per_page + 1),
            Group.without_key(Spec(field="items", spec=Push("$$ROOT"))),
            Project(
                Spec.exclude("_id"),
                Spec.include("items"),
            ),
        ).expression()



from .consumer import MessageConsumer
from .consumer_runner import FastAPIMessageConsumerRunner, ThreadMessageConsumerRunner
from .event_bus import EventBus
from .model import EventMessage, SqsEventMessage
from .publisher import EventPublisher
from .queue.base import MessageQueue
from .queue.local import LocalMessageQueue, event_store
from .signal_handler import DefaultSignalHandler, SignalHandler, ThreadSignalHandler

__all__ = [
    "DefaultSignalHandler",
    "EventBus",
    "EventMessage",
    "EventPublisher",
    "FastAPIMessageConsumerRunner",
    "LocalMessageQueue",
    "MessageConsumer",
    "MessageQueue",
    "SignalHandler",
    "SqsEventMessage",
    "ThreadMessageConsumerRunner",
    "ThreadSignalHandler",
    "event_store",
]

try:
    from .queue.sqs import SqsMessageQueue  # noqa: F401

    __all__.append("SqsMessageQueue")
except ImportError:
    pass


try:
    from .consumer.fastapi import MessageConsumerASGIMiddleware  # noqa: F401

    __all__.append("MessageConsumerASGIMiddleware")
except ImportError:
    pass



from .. import utils

AEq = utils.create_general_aggregation_operator(name="AEq", mongo_operator="$eq")
ANe = utils.create_general_aggregation_operator(name="ANe", mongo_operator="$ne")
AGt = utils.create_general_aggregation_operator(name="AGt", mongo_operator="$gt")
AGte = utils.create_general_aggregation_operator(name="AGte", mongo_operator="$gte")
ALt = utils.create_general_aggregation_operator(name="ALt", mongo_operator="$lt")
ALte = utils.create_general_aggregation_operator(name="ALte", mongo_operator="$lte")



from __future__ import annotations

import io
from collections.abc import Callable, Generator
from typing import IO, TYPE_CHECKING, Annotated, Any, Self, TypeVar

from pydantic import GetPydanticSchema
from pydantic_core import core_schema

if TYPE_CHECKING:
    from .types import PyObjectId

IdT = TypeVar("IdT")

type Validator = Callable[[Any], Self]
type ValidatorGenerator = Generator[Validator, None, None]

type Factory[T] = Callable[..., T]
type PyObjectIdFactory = Factory[PyObjectId]

type NoArgsFactory[T] = Callable[[], T]
type NoArgsPyObjectIdFactory = NoArgsFactory[PyObjectId]


type BinaryIO = Annotated[
    io.BytesIO | IO[bytes],
    GetPydanticSchema(
        lambda tp, handler: core_schema.no_info_after_validator_function(
            validate_binary_io, core_schema.any_schema()
        )
    ),
]


def validate_binary_io[T](value: T) -> T:
    if not (isinstance(value, io.BytesIO) or isinstance(value, io.BufferedReader)):
        raise ValueError(
            f"Expected an instance of io.BytesIO or io.BufferedReader, got {value.__class__.__name__}"
        )
    return value



import types
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Self, cast

from compose import typing

from .helper import CoreSchemaGettable

MARKER_IS_COMPOSE_VALIDATOR = "_is_compose_validator"
MARKER_COMPOSE_VALIDATORS = "_compose_validators"


def caster[T](factory: Callable[[Any], T], /) -> Callable[[Any], T]:
    def _cast(v: Any) -> T:
        return factory(v)

    return _cast


def validator[**P, T](func: Callable[P, T]) -> Callable[P, T]:
    setattr(func, MARKER_IS_COMPOSE_VALIDATOR, True)
    return func


class ValidatablePrimitive[T]:
    if TYPE_CHECKING:

        def __init__(self, *args, **kwargs) -> None: ...

    @classmethod
    def __get_validators__(cls) -> typing.ValidatorGenerator:
        yield from getattr(cls, MARKER_COMPOSE_VALIDATORS, [])

    @classmethod
    @validator
    def cast(cls, v: T) -> Self:
        return cls(v)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        cls._compose_validators = [
            getattr(cls, name)
            for base_cls in cls.__mro__
            for name, member in base_cls.__dict__.items()
            if (
                isinstance(member, classmethod)
                and hasattr(member.__wrapped__, MARKER_IS_COMPOSE_VALIDATOR)
            )
        ]


class Str(str, ValidatablePrimitive[str], CoreSchemaGettable[str]): ...


class Int(int, ValidatablePrimitive[int], CoreSchemaGettable[int]): ...


class Float(float, ValidatablePrimitive[float], CoreSchemaGettable[float]): ...


def _create_list_type[T](t: type[T], /) -> type[list[T]]:
    def __get_validators__(c) -> typing.ValidatorGenerator:
        yield caster(c)

    return cast(
        type[list[T]],
        types.new_class(
            f"{t.__name__.title()}List",
            (list[t], ValidatablePrimitive[t], CoreSchemaGettable[list[t]]),
            exec_body=lambda ns: ns.update(
                {
                    "__get_validators__": classmethod(__get_validators__),
                }
            ),
        ),
    )


def create_list_type[T]() -> Callable[[type[T]], type[list[T]]]:
    cache = {}

    def factory(t: type[T]) -> type[list[T]]:
        type_name = t.__name__

        if (cached := cache.get(type_name)) is not None:
            return cached

        _result = _create_list_type(t)
        cache[type_name] = _result
        return _result

    return factory


TypedList = create_list_type()
StrList = TypedList(str)
IntList = TypedList(int)


class ListMeta(type):
    _cache = {}

    def __getitem__(self, item):
        type_name = item.__name__

        if (cached := self._cache.get(type_name)) is not None:
            return cached

        result = _create_list_type(item)
        self._cache[type_name] = result
        return result


class List[T](list[T], metaclass=ListMeta): ...



from typing import Generic

from .. import field, model, types
from ..typing import IdT


class Event(model.BaseModel, Generic[IdT]):
    id: IdT
    published_at: types.DateTime = field.DateTimeField()



from typing import Any

from .base import Operator, evaluate
from .types import DictExpression


class Expr(Operator):
    def __init__(self, op: Any):
        self.op = op

    def expression(self) -> DictExpression:
        return {"$expr": evaluate(self.op)}



from collections.abc import Callable
from typing import Any

import pendulum
from pydantic import Field

from . import types


class _PyObjectIdField:
    def __call__(self, **kwargs) -> Any:
        default_kwargs = {"alias": "_id"}
        return Field(**(default_kwargs | kwargs))  # type: ignore


class _DatetimeField:
    def __call__(self, **kwargs) -> types.DateTime:
        default_kwargs = {"default_factory": pendulum.DateTime.utcnow}
        return Field(**(default_kwargs | kwargs))  # type: ignore


PyObjectIdField: Callable[..., Any] = _PyObjectIdField()
DateTimeField: Callable[..., types.DateTime] = _DatetimeField()



from __future__ import annotations

import functools
import uuid
from typing import Any, ClassVar, get_args, get_origin

import pendulum
from sqlalchemy import Table, delete, insert, select, update
from sqlalchemy.orm import Session

from ..entity import SQLEntity
from .base import BaseRepository


class SQLRepository[T: SQLEntity](BaseRepository):
    __table__: ClassVar[Table]

    def find_by_id(self, entity_id: uuid.UUID, session: Session) -> T | None:
        return self.find_by({"id": str(entity_id)}, session=session)

    def find_by(self, filter_: dict[str, Any], session: Session) -> T | None:
        table = self.__table__
        stmt = select(table).filter_by(**filter_)
        row = session.execute(stmt).mappings().first()
        if row is None:
            return None

        return self._entity_type.model_validate(dict(row))

    def list_by(
        self,
        filter_: dict[str, Any],
        session: Session,
        *,
        sort: list[tuple[str, int]] | None = None,
    ) -> list[T]:
        table = self.__table__
        stmt = select(table).filter_by(**filter_)

        if sort is not None:
            order_clauses = [
                table.c[col].asc() if direction == 1 else table.c[col].desc()
                for col, direction in sort
            ]
            stmt = stmt.order_by(*order_clauses)

        rows = session.execute(stmt).mappings().all()
        return [self._entity_type.model_validate(dict(row)) for row in rows]

    def add(self, entity: T, session: Session) -> None:
        table = self.__table__
        row = self._to_row(entity)
        session.execute(insert(table).values(**row))

    def add_many(self, entities: list[T], session: Session) -> None:
        table = self.__table__
        rows = [self._to_row(entity) for entity in entities]
        session.execute(insert(table), rows)

    def update(self, entity: T, session: Session) -> None:
        table = self.__table__
        row = entity.model_dump(mode="json", exclude={"id"})

        if "updated_at" in row:
            now = pendulum.DateTime.utcnow()
            entity.updated_at = now
            row["updated_at"] = now.isoformat()

        session.execute(update(table).where(table.c.id == str(entity.id)).values(**row))

    def delete(self, entity_id: uuid.UUID, session: Session) -> None:
        table = self.__table__
        session.execute(delete(table).where(table.c.id == str(entity_id)))

    @functools.cached_property
    def _entity_type(self) -> type[T]:
        orig_base = next(
            (base for base in self.__class__.__orig_bases__ if get_origin(base) is SQLRepository),
            None,
        )
        if orig_base is None:
            raise ValueError(
                f"{self.__class__.__name__} must inherit SQLRepository[T] with explicit type parameter"
            )

        return get_args(orig_base)[0]

    @staticmethod
    def _to_row(entity: T) -> dict[str, Any]:
        return entity.model_dump(mode="json")



from __future__ import annotations

import abc
import functools
import operator
from collections.abc import Callable
from typing import Any, ClassVar, Self

from .types import DictExpression, ListExpression


class Operator(abc.ABC):
    @abc.abstractmethod
    def expression(self) -> Any:
        raise NotImplementedError

    def pipe(self, op: Callable[..., Operator]) -> Operator:
        return op(self.expression())


class ComparisonOperator(Operator):
    def __init__(self, field: str, value: Any | None = None):
        self.field = field
        self.value = value

    @abc.abstractmethod
    def expression(self) -> dict[str, Any]:
        raise NotImplementedError


class EqualityOperator(ComparisonOperator):
    @abc.abstractmethod
    def expression(self) -> dict[str, Any]:
        raise NotImplementedError

    @classmethod
    def is_null(cls, field: str) -> Self:
        return cls(field=field, value=None)

    @classmethod
    def is_true(cls, field: str) -> Self:
        return cls(field=field, value=True)

    @classmethod
    def is_false(cls, field: str) -> Self:
        return cls(field=field, value=False)


class LogicalOperator(Operator):
    def __init__(self, *ops: Operator):
        self.ops = list(ops)

    @abc.abstractmethod
    def expression(self) -> dict[str, ListExpression]:
        raise NotImplementedError


class GeneralAggregationOperator(Operator):
    mongo_operator: ClassVar[str] = ""

    def __init__(self, *expressions: Any):
        self.expressions = list(expressions)

    def expression(self) -> DictExpression:
        return deep_evaluate({self.mongo_operator: self.expressions})


class Stage[T](Operator):
    @abc.abstractmethod
    def expression(self) -> T:
        raise NotImplementedError


class Merge[T](Operator):
    def __init__(self, *ops: Operator, initial: Any):
        self.ops = list(ops)
        self.initial = initial

    def expression(self) -> T:
        return functools.reduce(operator.or_, [op.expression() for op in self.ops], self.initial)

    @classmethod
    def dict(cls, *ops: *tuple[Operator, ...]) -> Merge[dict[str, Any]]:
        return cls(*ops, initial={})


def evaluate(value: Any) -> Any:
    if isinstance(value, Operator):
        return value.expression()
    return value


def deep_evaluate(value: Any) -> Any:
    match value:
        case Operator():
            return value.expression()
        case dict():
            return {k: deep_evaluate(v) for k, v in value.items()}
        case list():
            return [deep_evaluate(v) for v in value]
        case _:
            return value



from __future__ import annotations

import dataclasses
import enum
import logging
import os
import sys
from collections.abc import Iterable
from types import FrameType
from typing import TYPE_CHECKING, Protocol, Self, Unpack

try:
    from loguru import logger
except ImportError:
    raise ImportError("Install `loguru` extra to use logging features") from None

if TYPE_CHECKING:
    from loguru import BasicHandlerConfig, Logger, Record


class InterceptHandler(logging.Handler):
    """표준 logging으로 발생한 로그를 loguru로 전달하는 브릿지 핸들러.

    gunicorn, uvicorn처럼 자체 핸들러를 등록하는 라이브러리는 root 로거의 loguru 설정이
    덮어써지므로, 해당 로거의 핸들러를 직접 교체하여 loguru 파이프라인으로 통합함.

    https://loguru.readthedocs.io/en/stable/overview.html#entirely-compatible-with-standard-logging
    """

    def emit(self, record: logging.LogRecord) -> None:
        # 콜스택을 역추적하여 실제 로그 호출 지점의 depth를 계산함.
        # loguru가 InterceptHandler 내부가 아닌 원래 로그 발생 위치를 기록하기 위함.
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        frame: FrameType | None = logging.currentframe()
        depth = 0
        while frame is not None and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


def route_to_loguru(
    intercept_handler: InterceptHandler,
    logger_names: Iterable[str] = (
        "gunicorn.error",
        "gunicorn.access",
        "uvicorn.error",
        "uvicorn.access",
    ),
) -> None:
    logging.basicConfig(handlers=[intercept_handler], level=logging.INFO, force=True)
    for name in logger_names:
        lg = logging.getLogger(name)
        lg.handlers = [intercept_handler]
        lg.propagate = False


class LogFilterOp(Protocol):
    def __call__(self, record: Record) -> bool: ...


class LogFilterNotContains:
    def __init__(self, pattern: str):
        self.pattern = pattern

    def __call__(self, record: Record) -> bool:
        return self.pattern not in record["message"]


class LogFilter:
    def __init__(self, *ops: *tuple[LogFilterOp, ...]):
        self.ops = list(ops)

    def __call__(self, record: Record) -> bool:
        return all(op(record) for op in self.ops)


class LogFormat(enum.StrEnum):
    SERIALIZED = "{message}"
    NON_SERIALIZED = (
        "{time:YYYY-MM-DD HH:mm:ss.SSS} | <level>{level: <8}</level> | <level>{message}</level>"
    )


@dataclasses.dataclass
class LogDisplayConfig:
    format: str
    colorize: bool
    serialize: bool

    @classmethod
    def serialized(cls) -> Self:
        return cls(
            format=LogFormat.SERIALIZED,
            colorize=False,
            serialize=True,
        )

    @classmethod
    def non_serialized(cls) -> Self:
        return cls(
            format=LogFormat.NON_SERIALIZED,
            colorize=True,
            serialize=False,
        )


def create_logger(**config: Unpack[BasicHandlerConfig]) -> Logger:
    level = config.get("level", logging.INFO)

    intercept_handler = InterceptHandler()
    logging.basicConfig(handlers=[intercept_handler], level=level, force=True)
    route_to_loguru(intercept_handler=intercept_handler)

    logger.configure(
        handlers=[
            {
                "sink": sys.stdout,
                "level": level,
                "diagnose": False,
                "filter": LogFilter(
                    LogFilterNotContains("/health-check"),
                    LogFilterNotContains("/metrics"),
                ),
                **dataclasses.asdict(LogDisplayConfig.non_serialized()),
            }
            | config
        ]
    )

    return logger


class LogLevel(int):
    @classmethod
    def from_env(cls, env: str) -> Self:
        return cls(os.getenv(env, logging.INFO))



from typing import Protocol

from .. import model


class MessageConsumerType(Protocol):
    async def run(self) -> None: ...

    async def consume(self, message: model.EventMessage) -> None: ...

    def shutdown(self) -> None: ...



import functools
from collections.abc import Callable
from typing import Any

from fastapi import APIRouter as _FastAPIRouter
from fastapi.routing import APIRoute

from ..di.dependency_injector.wiring import Provider
from .wiring import auto_wired

_OVERRIDE_DEFAULTS = {"response_model_by_alias": False}
_OVERRIDE_METHODS = {
    "get",
    "post",
    "put",
    "delete",
    "options",
    "head",
    "patch",
    "trace",
    "api_route",
    "add_api_route",
}


def _with_default_overrides[**P, T](func: Callable[P, T]) -> Callable[P, T]:
    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        for key, value in _OVERRIDE_DEFAULTS.items():
            kwargs.setdefault(key, value)
        return func(*args, **kwargs)

    return wrapper


class APIRouter(_FastAPIRouter):
    pass


for _method_name in _OVERRIDE_METHODS:
    setattr(
        APIRouter,
        _method_name,
        _with_default_overrides(getattr(_FastAPIRouter, _method_name)),
    )


@functools.lru_cache(1)
def create_auto_wired_route(provider: Provider) -> type[APIRoute]:
    """
    Provider를 사용하는 AutoWiredAPIRoute를 생성하는 팩토리 함수.
    클래스 변수 대신 클로저를 사용하여 해시 문제를 회피
    """

    class AutoWiredAPIRoute(APIRoute):
        def __init__(self, path: str, endpoint: Callable[..., Any], **kwargs: Any):
            super().__init__(
                path=path,
                endpoint=auto_wired(provider)(endpoint),
                **kwargs,
            )

    return AutoWiredAPIRoute



import functools
import os
import sys
import time
import uuid
from collections.abc import Callable, Generator
from typing import Any, get_type_hints


def descendants_of[T](cls: type[T]) -> Generator[type[T], None, None]:
    stack = cls.__subclasses__()
    while stack:
        current_cls = stack.pop()
        yield current_cls
        stack.extend(current_cls.__subclasses__())


def unordered_partial[RT, T](p: functools.partial[RT], t: T) -> Callable[..., RT]:
    type_hints = get_type_hints(p.func)

    exclude_keys = {*p.keywords, "return"}
    candidates = [k for k, v in type_hints.items() if v == t and k not in exclude_keys]

    if len(candidates) != 1:
        raise TypeError(
            f"Cannot inject argument of type {t} into {p}. "
            f"Expected exactly one argument of type {t}, "
            f"but found {len(candidates)}: {candidates}"
        )

    arg_name = candidates[0]

    def wrapper(arg: Any) -> RT:
        return p(**{arg_name: arg})

    return wrapper


def uuid4_hex() -> str:
    return uuid.uuid4().hex


if sys.version_info >= (3, 14):

    def uuid7() -> uuid.UUID:
        return uuid.uuid7()

else:

    def uuid7() -> uuid.UUID:
        timestamp_ms = int(time.time() * 1000)
        rand = int.from_bytes(os.urandom(10))
        value = (timestamp_ms & 0xFFFF_FFFF_FFFF) << 80 | rand
        value = (value & ~(0xF << 76)) | (0x7 << 76)
        value = (value & ~(0x3 << 62)) | (0x2 << 62)
        return uuid.UUID(int=value)


def uuid7_str() -> str:
    return str(uuid7())



from collections.abc import Iterator


def find[T](iterator: Iterator[T], /) -> T | None:
    return next(iterator, None)


def unwrap[T](v: T | None, exc: Exception, /) -> T:
    if v is None:
        raise exc
    return v


def ident[T](x: T) -> T:
    return x



from __future__ import annotations

from typing import Any, TypeVar

DictExpression = dict[str, Any]
ListExpression = list[DictExpression]
Expression = TypeVar("Expression", DictExpression, ListExpression)

Id = "$_id"


class _FieldPath(str):
    def __new__(cls, v: str):
        if not v.startswith("$"):
            raise ValueError("path must be prefixed with $")

        return super().__new__(cls, v)


class _String(str):
    def __new__(cls, v: str):
        if v.startswith("$"):
            raise ValueError("string must not be prefixed with $")

        return super().__new__(cls, v)


class _NonNegativeInt(int):
    def __new__(cls, v: int):
        if v < 0:
            raise ValueError(f"Expected non-negative integer, got {v}")
        return super().__new__(cls, v)


class _PositiveInt(int):
    def __new__(cls, v: int):
        if v <= 0:
            raise ValueError(f"Expected positive integer, got {v}")
        return super().__new__(cls, v)



from collections.abc import Callable
from typing import Any, cast

from .base import GeneralAggregationOperator, Operator


def create_operator[T: Operator](
    name: str,
    expression_factory: Callable[[T], Any],
    base: tuple[type[T], ...],
) -> type[T]:
    return cast(
        type[T],
        type(name, base, {"expression": expression_factory}),
    )


def create_general_aggregation_operator(
    name: str, mongo_operator: str
) -> type[GeneralAggregationOperator]:
    return cast(
        type[GeneralAggregationOperator],
        type(name, (GeneralAggregationOperator,), {"mongo_operator": mongo_operator}),
    )



from compose.di.dependency_injector.wiring import (
    DEFAULT_RESOLVABLE_PROVIDER_TYPES,
    Provider,
    create_provider,
    create_resolver,
    get_wiring_packages,
    provide,
    resolve,
    resolve_by_name,
    resolve_by_object_name,
)

__all__ = [
    "DEFAULT_RESOLVABLE_PROVIDER_TYPES",
    "Provider",
    "create_provider",
    "create_resolver",
    "get_wiring_packages",
    "provide",
    "resolve",
    "resolve_by_name",
    "resolve_by_object_name",
]



import asyncio
import collections
import contextlib

from ..event_bus import EventBus
from ..model import EventMessage
from ..queue.local import LocalMessageQueue, event_store
from .hook import DEFAULT_HOOKS, Hook, HookArgType, HookEventType, default_hook

try:
    from starlette.types import ASGIApp, Receive, Scope, Send
except ImportError:
    raise ImportError("Please install `starlette` to use `MessageConsumerASGIMiddleware`")


class MessageConsumerASGIMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        event_bus: EventBus,
        message_queue: LocalMessageQueue,
        hooks: dict[HookEventType, list[Hook]] | None = None,
    ) -> None:
        self.app = app
        self.event_bus = event_bus
        self.message_queue = message_queue
        self.hooks = DEFAULT_HOOKS | (hooks or {})

        self._default_hook = default_hook

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):
            await self.app(scope, receive, send)
            return

        self._execute_hook("on_start", "MessageConsumer started")
        with self.event_store_context():
            try:
                await self.app(scope, receive, send)
            finally:
                await self.consume_messages(event_store.get())
                self._execute_hook("on_shutdown", "MessageConsumer shutting down")

    @contextlib.contextmanager
    def event_store_context(self):
        token = event_store.set(collections.deque())

        try:
            yield
        finally:
            event_store.reset(token)

    async def consume_messages(self, messages: collections.deque[EventMessage]) -> None:
        await asyncio.gather(*(self.consume(message) for message in messages))

    async def consume(self, message: EventMessage) -> None:
        self._execute_hook("on_receive", message)
        try:
            await self.event_bus.handle_event(message.body)
            self.message_queue.delete(message)
        except Exception as exc:
            self._execute_hook("on_consume_error", exc)
            return

        self._execute_hook("on_consume", message)

    def _execute_hook(self, hook_event_type: HookEventType, arg: HookArgType, /) -> None:
        for hook in self.hooks.get(hook_event_type, [self._default_hook]):
            hook(arg)



from .base import Query

__all__ = ["Query"]

try:
    from .mongo.query import (
        MongoCursorPaginationQuery,
        MongoOffsetPaginationQuery,
        MongoPaginationQuery,
        MongoQuery,
    )

    __all__ += [
        "MongoCursorPaginationQuery",
        "MongoOffsetPaginationQuery",
        "MongoPaginationQuery",
        "MongoQuery",
    ]
except ImportError:
    pass



from __future__ import annotations

import urllib.parse
from typing import Any, ClassVar, cast

from .primitive import Str


class _S3ObjectUrl(Str):
    base_url: ClassVar[str]

    def __new__(cls, v: Any):
        if getattr(cls, "base_url", None) is None:
            raise ValueError(
                f"`base_url` must be set on `{cls.__name__}`. "
                f"Use `create_s3_object_url()` to create a configured type"
            )

        key = _extract_key(str(v), base_url=cls.base_url)
        encoded_path = _encode_path(key)
        return super().__new__(cls, f"{cls.base_url}/{encoded_path}")


def create_s3_object_url(base_url: str) -> type[_S3ObjectUrl]:
    return cast(
        type[_S3ObjectUrl],
        type("S3ObjectUrl", (_S3ObjectUrl,), {"base_url": base_url}),
    )


def _extract_key(v: str, *, base_url: str) -> str:
    v = urllib.parse.unquote(v.strip("/"))
    return v.removeprefix(base_url).lstrip("/")


def _encode_path(key: str) -> str:
    return "/".join(urllib.parse.quote(part, safe="~()*!.'") for part in key.split("/"))



from typing import Generic

from pydantic import ConfigDict

from .. import model, schema
from ..typing import IdT


class Command(model.BaseModel): ...


class UserCommand(Command, Generic[IdT]):
    user_id: IdT | None = None

    model_config = ConfigDict(json_schema_extra=schema.extra.schema_excludes("user_id"))



from typing import Self

from .endpoint import SpecialEndpoint

_registry = set()


class NonInstrumentedUrls(list[str]):
    @classmethod
    def register(cls, *urls: *tuple[str, ...]) -> None:
        for url in urls:
            _registry.add(url)

    @classmethod
    def current(cls) -> Self:
        return cls(_registry)


NonInstrumentedUrls.register(*SpecialEndpoint)



from typing import Any

from pydantic import SerializeAsAny, SkipValidation

from compose import model
from compose.event import Event


# body 필드의 타입 선언에 대한 배경:
#
# 이전에는 TypeVar("MessageBody", bound=Event)를 사용함. Generic을 상속하지 않는
# 클래스에서 TypeVar는 타입 체커에서 의미가 없고(pyrefly bad-argument-type 에러),
# Pydantic은 미해결 TypeVar를 Any로 취급하여 검증과 직렬화를 모두 건너뜀.
# 즉, 타입 안전성 없이 런타임에서만 우연히 동작하는 상태였음.
#
# 단순히 Event나 Event[Any]로 바꾸면 Pydantic이 Event 스키마로 검증하면서
# 하위 클래스 인스턴스를 Event로 변환하고(extra="ignore"로 추가 필드 제거),
# 직렬화 시에도 Event 스키마를 사용하여 하위 클래스의 커스텀 타입(PyObjectId 등)을
# 처리하지 못함.
#
# SkipValidation: 검증을 건너뛰어 하위 클래스 인스턴스를 그대로 보존
# SerializeAsAny: 선언 타입이 아닌 런타임 실제 타입의 직렬화기를 사용
class EventMessage(model.BaseModel):
    body: SerializeAsAny[SkipValidation[Event[Any]]]


class SqsEventMessage(EventMessage):
    receipt_handle: str



from typing import Generic

from .base import Operator, deep_evaluate
from .types import Expression


class Raw(Generic[Expression], Operator):
    def __init__(self, expression: Expression):
        self._expression = expression  # type: ignore

    def expression(self) -> Expression:
        return deep_evaluate(self._expression)



from collections.abc import Iterable
from typing import Self

from dependency_injector import containers


class DeclarativeContainer(containers.DeclarativeContainer):
    @classmethod
    def wired(
        cls,
        modules: Iterable[str] | None = None,
        packages: Iterable[str] = (),
        from_package: str | None = None,
        warn_unresolved: bool = False,
    ) -> Self:
        container = cls()
        container.check_dependencies()
        container.wire(
            modules=modules,
            packages=packages,
            from_package=from_package,
            warn_unresolved=warn_unresolved,
        )
        return container



from typing import Any, Self

from .base import Merge, Operator, Stage, deep_evaluate
from .evaluation import Expr
from .logical import And, Nor, Or
from .pipeline import Pipeline
from .sort import SortBy
from .types import DictExpression, _FieldPath, _NonNegativeInt, _PositiveInt


class EmptyStage(Stage[DictExpression]):
    def expression(self) -> DictExpression:
        return {}


class Match(Stage[DictExpression]):
    def __init__(self, op: Operator):
        self.op = op

    def expression(self) -> DictExpression:
        if not (expression := self.op.expression()):
            return EmptyStage().expression()
        return {"$match": expression}

    @classmethod
    def and_(cls, *ops: Operator) -> Self:
        return cls(And(*ops))

    @classmethod
    def or_(cls, *ops: Operator) -> Self:
        return cls(Or(*ops))

    @classmethod
    def nor(cls, *ops: Operator) -> Self:
        return cls(Nor(*ops))

    @classmethod
    def expr(cls, op: Any) -> Self:
        return cls(Expr(op))


class Sort(Stage[DictExpression]):
    def __init__(self, *criteria: SortBy):
        self.criteria = list(criteria)

    def expression(self) -> DictExpression:
        if not (merged := Merge.dict(*self.criteria).expression()):
            raise ValueError("Sort requires at least one SortBy criterion")
        return {"$sort": merged}


class Spec(Operator):
    def __init__(self, field: str, spec: Any):
        self.field = field
        self.spec = spec

    def expression(self) -> DictExpression:
        return {self.field: deep_evaluate(self.spec)}

    @classmethod
    def include(cls, field: str) -> Self:
        return cls(field=field, spec=1)

    @classmethod
    def exclude(cls, field: str) -> Self:
        return cls(field=field, spec=0)

    @classmethod
    def alias(cls, field: str, spec: str) -> Self:
        return cls(field=field, spec=_FieldPath(spec))


class Project(Stage[DictExpression]):
    def __init__(self, *specs: Spec):
        self.specs = list(specs)

    def expression(self) -> DictExpression:
        return {"$project": Merge.dict(*self.specs).expression()}

    @classmethod
    def of(cls, **attrs: Any) -> Self:
        return cls(*(Spec(field, spec) for field, spec in attrs.items()))


class Lookup(Stage[DictExpression]):
    def __init__(self, from_: str, as_: str | None = None):
        self.from_ = from_
        self.as_ = as_ or self.from_


class EqualityLookup(Lookup):
    def __init__(self, from_: str, local_field: str, foreign_field: str, as_: str | None = None):
        super().__init__(from_=from_, as_=as_)
        self.local_field = local_field
        self.foreign_field = foreign_field

    def expression(self) -> DictExpression:
        return {
            "$lookup": {
                "from": self.from_,
                "as": self.as_,
                "localField": self.local_field,
                "foreignField": self.foreign_field,
            }
        }

    @classmethod
    def on_id(cls, from_: str, local_field: str, as_: str | None = None) -> Self:
        return cls(
            from_=from_,
            local_field=local_field,
            foreign_field="_id",
            as_=as_,
        )


class SubqueryLookup(Lookup):
    def __init__(
        self,
        from_: str,
        pipeline: Pipeline,
        let: Operator | None = None,
        as_: str | None = None,
    ):
        super().__init__(from_=from_, as_=as_)
        self.let = let
        self.pipeline = pipeline

    def expression(self) -> DictExpression:
        result: DictExpression = {
            "from": self.from_,
            "as": self.as_,
            "pipeline": self.pipeline.expression(),
        }
        if self.let is not None:
            result["let"] = self.let.expression()
        return {"$lookup": result}


class Unwind(Stage[DictExpression]):
    def __init__(
        self,
        path: str,
        include_array_index: str | None = None,
        preserve_null_and_empty_arrays: bool | None = None,
    ):
        self.path = _FieldPath(path)
        self.include_array_index = include_array_index
        self.preserve_null_and_empty_arrays = preserve_null_and_empty_arrays

    def expression(self) -> DictExpression:
        result: DictExpression = {"path": self.path}
        if self.include_array_index is not None:
            result["includeArrayIndex"] = self.include_array_index
        if self.preserve_null_and_empty_arrays is not None:
            result["preserveNullAndEmptyArrays"] = self.preserve_null_and_empty_arrays
        return {"$unwind": result}

    @classmethod
    def preserve_missing(cls, path: str, include_array_index: str | None = None) -> Self:
        return cls(
            path=path,
            include_array_index=include_array_index,
            preserve_null_and_empty_arrays=True,
        )


class Set(Stage[DictExpression]):
    def __init__(self, *specs: Spec):
        self.specs = list(specs)

    def expression(self) -> DictExpression:
        return {"$set": Merge.dict(*self.specs).expression()}


class FacetSubPipeline(Operator):
    def __init__(self, output_field: str, pipeline: Pipeline):
        self.output_field = output_field
        self.pipeline = pipeline

    def expression(self) -> DictExpression:
        return {self.output_field: self.pipeline.expression()}


class Facet(Stage[DictExpression]):
    def __init__(self, *pipelines: FacetSubPipeline):
        self.pipelines = list(pipelines)

    def expression(self) -> DictExpression:
        return {"$facet": Merge.dict(*self.pipelines).expression()}


class Skip(Stage[DictExpression]):
    def __init__(self, skip: int):
        self.skip = _NonNegativeInt(skip)

    def expression(self) -> DictExpression:
        return {"$skip": self.skip}


class Limit(Stage[DictExpression]):
    def __init__(self, limit: int):
        self.limit = _PositiveInt(limit)

    def expression(self) -> DictExpression:
        return {"$limit": self.limit}


class TextSearchOperator(Operator):
    def __init__(self, query: str, path: str | list[str]):
        self.query = query
        self.path = path

    def expression(self) -> DictExpression:
        return {"text": {"query": self.query, "path": self.path}}


class Search(Stage[DictExpression]):
    def __init__(self, index: str, op: Operator):
        self.index = index
        self.op = op

    def expression(self) -> DictExpression:
        return {"$search": {"index": self.index} | self.op.expression()}


class OffsetPagination(Stage[DictExpression]):
    def __init__(
        self,
        page: int,
        per_page: int,
        metadata_ops: list[Operator] | None = None,
    ):
        self.page = _PositiveInt(page)
        self.per_page = _PositiveInt(per_page)
        self.metadata_ops = metadata_ops or []

    def expression(self) -> DictExpression:
        return {
            "$facet": {
                "metadata": [{"$count": "total"}, *[op.expression() for op in self.metadata_ops]],
                "items": [
                    Skip((self.page - 1) * self.per_page).expression(),
                    Limit(self.per_page).expression(),
                ],
            }
        }


class ReplaceRoot(Stage[DictExpression]):
    def __init__(self, new_root: Any):
        self.new_root = new_root

    def expression(self) -> DictExpression:
        return {"$replaceRoot": {"newRoot": deep_evaluate(self.new_root)}}


class Group(Stage[DictExpression]):
    def __init__(self, *ops: Operator, key: Any | None = None):
        self.ops = list(ops)
        self.key = key

    def expression(self) -> DictExpression:
        return {
            "$group": {
                "_id": deep_evaluate(self.key),
                **Merge.dict(*self.ops).expression(),
            }
        }

    @classmethod
    def without_key(cls, *ops: Operator) -> Self:
        return cls(*ops, key=None)


class Sample(Stage[DictExpression]):
    def __init__(self, size: int):
        self.size = _NonNegativeInt(size)

    def expression(self) -> DictExpression:
        return {"$sample": {"size": self.size}}



import os
from typing import Any

from pydantic import Field

from compose.model import BaseModel


class GunicornSettings(BaseModel):
    wsgi_app: str
    bind: str = "0.0.0.0:80"
    workers: int = Field(default_factory=lambda: (os.cpu_count() or 1) + 1)
    worker_class: str = "uvicorn_worker.UvicornWorker"
    timeout: int = 120
    max_requests: int | None = None
    max_requests_jitter: int | None = None


def export_settings(
    globals_: dict[str, Any],
    settings: GunicornSettings,
    *,
    env: str | None = None,
    overrides: dict[str, dict[str, Any]] | None = None,
    **kwargs: Any,
) -> None:
    overrides = overrides or {}

    if env is not None and (update := overrides.get(env, {})):
        settings = settings.model_copy(update=update)

    globals_ |= settings.model_dump(exclude_none=True) | kwargs



import asyncio
from collections.abc import Awaitable, Callable, Hashable


class AsyncJob[K: Hashable, T]:
    def __init__[**P](
        self,
        key: K,
        func: Callable[P, Awaitable[T]],
        *args: P.args,
        **kwargs: P.kwargs,
    ):
        self.key = key
        self._invoke: Callable[[], Awaitable[T]] = lambda: func(*args, **kwargs)

    def __call__(self) -> Awaitable[T]:
        return self._invoke()


class AsyncTaskExecutor:
    def __init__(self, concurrency: int, timeout: float | None = None):
        self.concurrency = concurrency
        self.timeout = timeout

    async def execute[K, T](
        self,
        jobs: list[AsyncJob[K, T]],
        concurrency: int | None = None,
        timeout: float | None = None,
    ) -> dict[K, T]:
        concurrency = concurrency if concurrency is not None else self.concurrency
        timeout = timeout if timeout is not None else self.timeout

        coro = _execute_jobs(jobs=jobs, semaphore=asyncio.Semaphore(concurrency))

        if timeout is None:
            return await coro

        async with asyncio.timeout(timeout):
            return await coro


async def _execute_jobs[K, T](
    jobs: list[AsyncJob[K, T]],
    semaphore: asyncio.Semaphore,
) -> dict[K, T]:
    try:
        async with asyncio.TaskGroup() as tg:
            tasks = [
                (job.key, tg.create_task(_execute_job(job=job, semaphore=semaphore)))
                for job in jobs
            ]
    except* Exception as eg:
        raise eg.exceptions[0]

    return {key: task.result() for key, task in tasks}


async def _execute_job[K, T](job: AsyncJob[K, T], semaphore: asyncio.Semaphore) -> T:
    async with semaphore:
        return await job()



from typing import Self

import pendulum
from authlib.jose import errors, jwt

from .. import exceptions
from . import vo

_STANDARD_CLAIM_KEYS = frozenset({"sub", "iss", "jti", "iat", "exp"})


class JWTDecoder:
    def __init__(self, secret_key: str, clock: type[pendulum.DateTime]):
        self.secret_key = secret_key
        self.clock = clock

    @classmethod
    def default(cls, secret_key: str) -> Self:
        return cls(secret_key=secret_key, clock=pendulum.DateTime)

    def decode(self, token: str) -> vo.TokenClaims:
        try:
            decoded = jwt.decode(token.encode(), key=self.secret_key)
        except (errors.DecodeError, errors.BadSignatureError):
            raise exceptions.AuthorizationError("Invalid token")

        try:
            decoded.validate(now=int(self.clock.utcnow().timestamp()))
        except errors.ExpiredTokenError:
            raise exceptions.AuthorizationError("Token has expired")
        except errors.InvalidClaimError:
            raise exceptions.AuthorizationError("Invalid token claims")

        extra = {k: v for k, v in decoded.items() if k not in _STANDARD_CLAIM_KEYS}
        return vo.TokenClaims(
            sub=decoded["sub"],
            iss=decoded["iss"],
            jti=decoded["jti"],
            iat=decoded["iat"],
            exp=decoded["exp"],
            extra=extra,
        )



from typing import Literal, Self

import pymongo

from .base import Operator
from .types import DictExpression


class SortBy(Operator):
    def __init__(self, field: str, direction: Literal[1, -1]):
        self.field = field
        self.direction = direction

    def expression(self) -> DictExpression:
        return {self.field: self.direction}

    @classmethod
    def asc(cls, field: str) -> Self:
        return cls(field=field, direction=pymongo.ASCENDING)

    @classmethod
    def desc(cls, field: str) -> Self:
        return cls(field=field, direction=pymongo.DESCENDING)

    @classmethod
    def parse(cls, key: str) -> Self:
        if key.startswith("-"):
            return cls.desc(field=key[1:])
        return cls.asc(field=key)



from . import func  # noqa: F401
from .aggregation import *  # noqa: F403
from .base import *  # noqa: F403
from .comparison import *  # noqa: F403
from .evaluation import *  # noqa: F403
from .logical import *  # noqa: F403
from .pagination import *  # noqa: F403
from .pipeline import *  # noqa: F403
from .raw import *  # noqa: F403
from .sort import *  # noqa: F403
from .stage import *  # noqa: F403
from .types import *  # noqa: F403



from typing import Any

from ..base import GeneralAggregationOperator, Operator, evaluate
from ..types import DictExpression


class Concat(GeneralAggregationOperator):
    mongo_operator = "$concat"


class Split(Operator):
    def __init__(self, expr: Any, delimiter: str):
        self.expr = expr
        self.delimiter = delimiter

    def expression(self) -> DictExpression:
        return {"$split": [evaluate(self.expr), self.delimiter]}



import mimetypes
import os
import urllib.parse
from typing import ClassVar, Self, TypedDict

from .primitive import Str


class ContentDisposition(str):
    @classmethod
    def attachment(cls, filename: str) -> Self:
        return cls(f"attachment; filename*=UTF-8''{urllib.parse.quote(filename)}")


class MimeTypeInfo(TypedDict):
    type: str
    ext: str


class MimeType(Str):
    """MIME Type을 표현하는 타입

    mimetypes 패키지에서 제공하지 않는 타입을 사용하려면
    어플리케이션 진입점에서 `MimeType.register([MimeTypeInfo(type="...", ext="..."), ...])`를 호출한 뒤
    `MimeType.guess(...)` 호출
    """

    default_type: ClassVar[str] = "binary/octet-stream"
    _known_types: ClassVar[list[MimeTypeInfo]] = [
        MimeTypeInfo(type="audio/midi", ext=".mid"),
        MimeTypeInfo(type="audio/midi", ext=".midi"),
    ]
    _is_known_mime_types_registered: ClassVar[bool] = False

    @classmethod
    def register_known_types(cls) -> None:
        if cls._is_known_mime_types_registered:
            return

        for t in cls._known_types:
            mimetypes.add_type(**t)

        cls._is_known_mime_types_registered = True

    @classmethod
    def register(cls, *types: *tuple[MimeTypeInfo, ...]) -> None:
        for t in [*types]:
            mimetypes.add_type(**t)

    @classmethod
    def guess(cls, url: str | os.PathLike[str]) -> Self:
        mime_type, _ = mimetypes.guess_type(url)
        return cls(mime_type or cls.default_type)

    @classmethod
    def default(cls) -> Self:
        return cls(cls.default_type)


MimeType.register_known_types()



import pytest

from compose import enums

from .. import hook


@pytest.hookimpl(tryfirst=True)
def pytest_configure(config: pytest.Config) -> None:
    hook.check_env_is_allowed(enums.AppEnv.current())
    hook.register_test_type_markers(config)


@pytest.hookimpl()
def pytest_collection_modifyitems(items: list[pytest.Function]) -> None:
    hook.add_test_type_markers(items)



import enum
import http

from fastapi import FastAPI, Response


class SpecialEndpoint(enum.StrEnum):
    HEALTH_CHECK = "/health-check"
    METRICS = "/metrics"


def health_check() -> Response:
    return Response(status_code=http.HTTPStatus.OK)


def add_health_check_endpoint(app: FastAPI) -> None:
    app.get(SpecialEndpoint.HEALTH_CHECK, include_in_schema=False)(health_check)



from __future__ import annotations

from collections.abc import Callable, Collection
from typing import TYPE_CHECKING, Any

import wrapt
from loguru import logger
from opentelemetry import trace
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import unwrap

ORIGINAL_PATCHER_ATTR = "_otel_original_patcher"
_instruments = tuple()

if TYPE_CHECKING:
    from loguru import Logger, Record


RecordPatcher = Callable[["Record"], None]


def create_trace_patcher(tracer_provider: trace.TracerProvider) -> RecordPatcher:
    resource = getattr(tracer_provider, "resource", None)
    service_name = resource.attributes.get("service.name", "") if resource is not None else ""

    def patcher(record: Record) -> None:
        span = trace.get_current_span()
        ctx = span.get_span_context()
        is_valid_span = span != trace.INVALID_SPAN

        record["extra"].update(
            {
                "otel_service_name": service_name,
                "otel_trace_id": (
                    trace.format_trace_id(ctx.trace_id)
                    if is_valid_span
                    else str(trace.INVALID_TRACE_ID)
                ),
                "otel_span_id": (
                    trace.format_span_id(ctx.span_id)
                    if is_valid_span
                    else str(trace.INVALID_SPAN_ID)
                ),
                "otel_trace_sampled": ctx.trace_flags.sampled if is_valid_span else False,
            }
        )

    return patcher


def create_configure_wrapper(
    trace_patcher: RecordPatcher,
) -> Callable[..., list[int]]:
    def wrapped_configure(
        func: Callable[..., list[int]],
        instance: Logger,
        args: tuple[Any],
        kwargs: dict[str, Any],
    ) -> list[int]:
        original_patcher = kwargs.get("patcher")
        setattr(instance, ORIGINAL_PATCHER_ATTR, original_patcher)

        if original_patcher is None:
            return func(*args, **kwargs)

        user_patcher: RecordPatcher = original_patcher

        def wrapped_patcher(record: Record) -> None:
            user_patcher(record)
            trace_patcher(record)

        kwargs["patcher"] = wrapped_patcher
        return func(*args, **kwargs)

    return wrapped_configure


def _default_record_patcher(_: Record) -> None:
    return None


class LoguruInstrumentor(BaseInstrumentor):
    """
    Reference:
        https://github.com/open-telemetry/opentelemetry-python-contrib/tree/main/instrumentation/opentelemetry-instrumentation-logging
    """

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs) -> None:
        tracer_provider = kwargs.get("tracer_provider", trace.get_tracer_provider())

        trace_patcher = create_trace_patcher(tracer_provider)
        logger.configure(patcher=trace_patcher)
        wrapt.wrap_function_wrapper(
            logger,
            "configure",
            create_configure_wrapper(trace_patcher),
        )

    def _uninstrument(self, **kwargs) -> None:
        original_patcher = getattr(logger, ORIGINAL_PATCHER_ATTR, None)
        unwrap(logger, "configure")
        logger.configure(patcher=original_patcher or _default_record_patcher)
        if hasattr(logger, ORIGINAL_PATCHER_ATTR):
            delattr(logger, ORIGINAL_PATCHER_ATTR)



from typing import Any

from ..base import Operator, evaluate
from ..types import DictExpression


class ToString(Operator):
    def __init__(self, expr: Any):
        self.expr = expr

    def expression(self) -> DictExpression:
        return {"$toString": evaluate(self.expr)}


class ToBool(Operator):
    def __init__(self, expression: Any):
        self._expression = expression

    def expression(self) -> DictExpression:
        return {"$toBool": evaluate(self._expression)}


class ToInt(Operator):
    def __init__(self, expression: Any):
        self._expression = expression

    def expression(self) -> DictExpression:
        return {"$toInt": evaluate(self._expression)}



from .resource_server import ResourceServer
from .vo import AuthorizationGrant, TokenClaims, UserResource

__all__ = [
    "AuthorizationGrant",
    "ResourceServer",
    "TokenClaims",
    "UserResource",
]

try:
    from .authorization_server import AuthorizationServer
    from .token_decoder import JWTDecoder
    from .token_issuer import JWTIssuer

    __all__ += ["AuthorizationServer", "JWTDecoder", "JWTIssuer"]
except ImportError:
    pass

try:
    from .password import HashedPassword  # noqa: F401

    __all__.append("HashedPassword")
except ImportError:
    pass



import http
from collections.abc import Awaitable, Callable
from typing import ClassVar, Self

from fastapi import Request, Response
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse, PlainTextResponse
from pydantic import ValidationError
from starlette.exceptions import HTTPException

from ..exceptions import (
    AuthorizationError,
    DoesNotExistError,
    DomainValidationError,
    NotAllowedError,
)
from ..schema import Error, InvalidParam

type ExceptionHandler[E: Exception] = Callable[[Request, E], Response | Awaitable[Response]]


class ExceptionHandlerInfo:
    default_response_cls: ClassVar[type[Response]] = JSONResponse
    default_error_schema_cls: ClassVar[type[Error]] = Error

    def __init__(
        self, exc_class_or_status_code: int | type[Exception], handler: ExceptionHandler
    ) -> None:
        self.exc_class_or_status_code = exc_class_or_status_code
        self.handler = handler

    @classmethod
    def for_status_code(
        cls,
        status_code: http.HTTPStatus,
        error_type: str | None = None,
        response_cls: type[Response] | None = None,
        error_schema_cls: type[Error] | None = None,
    ) -> Self:
        return cls(
            exc_class_or_status_code=status_code,
            handler=create_exception_handler(
                status_code=status_code,
                error_type=error_type or status_code.name.lower(),
                response_cls=response_cls or cls.default_response_cls,
                error_schema_cls=error_schema_cls or cls.default_error_schema_cls,
            ),
        )

    @classmethod
    def for_unauthorized_error(
        cls,
        response_cls: type[Response] | None = None,
        error_schema_cls: type[Error] | None = None,
    ) -> Self:
        return cls(
            exc_class_or_status_code=http.HTTPStatus.UNAUTHORIZED,
            handler=create_unauthorized_error_handler(
                response_cls=response_cls or cls.default_response_cls,
                error_schema_cls=error_schema_cls or cls.default_error_schema_cls,
            ),
        )

    @classmethod
    def for_exc(
        cls,
        exc_cls: type[Exception],
        status_code: int,
        error_type: str | None = None,
        response_cls: type[Response] | None = None,
        error_schema_cls: type[Error] | None = None,
    ) -> Self:
        return cls(
            exc_class_or_status_code=exc_cls,
            handler=create_exception_handler(
                status_code=status_code,
                error_type=error_type or http.HTTPStatus(status_code).name.lower(),
                response_cls=response_cls or cls.default_response_cls,
                error_schema_cls=error_schema_cls or cls.default_error_schema_cls,
            ),
        )

    @classmethod
    def default(
        cls,
        error_schema_cls: type[Error] | None = None,
    ) -> Self:
        return cls(
            exc_class_or_status_code=Exception,
            handler=create_exception_handler(
                status_code=http.HTTPStatus.INTERNAL_SERVER_ERROR,
                error_type=http.HTTPStatus.INTERNAL_SERVER_ERROR.name.lower(),
                response_cls=cls.default_response_cls,
                error_schema_cls=error_schema_cls or cls.default_error_schema_cls,
            ),
        )

    @classmethod
    def for_http_exception(
        cls,
        response_cls: type[Response] | None = None,
        error_schema_cls: type[Error] | None = None,
    ) -> Self:
        return cls(
            exc_class_or_status_code=HTTPException,
            handler=create_http_exception_handler(
                response_cls=response_cls or cls.default_response_cls,
                error_schema_cls=error_schema_cls or cls.default_error_schema_cls,
            ),
        )

    @classmethod
    def for_request_validation_error(
        cls,
        exception_handler: ExceptionHandler | None = None,
        response_cls: type[Response] | None = None,
        error_schema_cls: type[Error] | None = None,
    ) -> Self:
        return cls(
            exc_class_or_status_code=RequestValidationError,
            handler=(
                exception_handler
                or create_default_validation_error_handler(
                    response_cls=response_cls or cls.default_response_cls,
                    error_schema_cls=error_schema_cls or cls.default_error_schema_cls,
                )
            ),
        )

    @classmethod
    def for_pydantic_validation_error(
        cls,
        exception_handler: ExceptionHandler | None = None,
        response_cls: type[Response] | None = None,
        error_schema_cls: type[Error] | None = None,
    ) -> Self:
        return cls(
            exc_class_or_status_code=ValidationError,
            handler=(
                exception_handler
                or create_default_validation_error_handler(
                    response_cls=response_cls or cls.default_response_cls,
                    error_schema_cls=error_schema_cls or cls.default_error_schema_cls,
                )
            ),
        )


def default_exception_handlers(
    *extras: *tuple[ExceptionHandlerInfo, ...],
    handler_info_cls: type[ExceptionHandlerInfo] = ExceptionHandlerInfo,
) -> tuple[ExceptionHandlerInfo, ...]:
    return (
        handler_info_cls.for_exc(DoesNotExistError, http.HTTPStatus.NOT_FOUND),
        handler_info_cls.for_exc(NotAllowedError, http.HTTPStatus.FORBIDDEN),
        handler_info_cls.for_exc(DomainValidationError, http.HTTPStatus.UNPROCESSABLE_ENTITY),
        handler_info_cls.for_exc(AuthorizationError, http.HTTPStatus.UNAUTHORIZED),
        *extras,
        handler_info_cls.for_request_validation_error(),
        handler_info_cls.for_pydantic_validation_error(),
        handler_info_cls.default(),
    )


def create_exception_handler(
    status_code: int,
    error_type: str,
    response_cls: type[Response],
    error_schema_cls: type[Error] = Error,
) -> ExceptionHandler:
    def exception_handler(request: Request, exc: Exception) -> Response:
        response = exc_to_error_schema(
            exc=exc, error_type=error_type, error_schema_cls=error_schema_cls
        )
        return response_cls(content=jsonable_encoder(response), status_code=status_code)

    return exception_handler


def create_unauthorized_error_handler(
    response_cls: type[Response],
    error_schema_cls: type[Error] = Error,
) -> ExceptionHandler:
    def exception_handler(request: Request, exc: Exception) -> Response:
        if (
            isinstance(exc, HTTPException)
            and exc.headers is not None
            and exc.headers.get("WWW-Authenticate", "").lower().startswith("basic")
        ):
            return PlainTextResponse(
                content=exc.detail,
                status_code=exc.status_code,
                headers=exc.headers,
            )

        response = exc_to_error_schema(
            exc=exc,
            error_type=(status_code := http.HTTPStatus.UNAUTHORIZED).name.lower(),
            error_schema_cls=error_schema_cls,
        )
        return response_cls(content=jsonable_encoder(response), status_code=status_code)

    return exception_handler


def exc_to_error_schema[T: Error](
    exc: Exception,
    error_type: str,
    error_schema_cls: type[T],
) -> T:
    return error_schema_cls(
        title=str(exc),
        type=error_type,
        detail=getattr(exc, "detail", None),
        invalid_params=(
            (invalid_params := getattr(exc, "invalid_params", None))
            and [InvalidParam.model_validate(obj=invalid_param) for invalid_param in invalid_params]
        ),
    )


def create_http_exception_handler(
    response_cls: type[Response],
    error_schema_cls: type[Error] = Error,
) -> ExceptionHandler:
    def http_exception_handler(request: Request, exc: HTTPException) -> Response:
        status_code = http.HTTPStatus(exc.status_code)
        return response_cls(
            content=jsonable_encoder(
                error_schema_cls(
                    title=exc.detail,
                    type=status_code.name.lower(),
                )
            ),
            status_code=status_code,
            headers=exc.headers,
        )

    return http_exception_handler


def create_default_validation_error_handler(
    response_cls: type[Response],
    error_schema_cls: type[Error] = Error,
) -> ExceptionHandler:
    def validation_error_handler(
        request: Request, exc: RequestValidationError | ValidationError
    ) -> Response:
        return response_cls(
            content=jsonable_encoder(
                error_schema_cls.from_validation_error(exc=exc, title="Validation failed"),
            ),
            status_code=http.HTTPStatus.UNPROCESSABLE_ENTITY,
        )

    return validation_error_handler



from __future__ import annotations

import functools
import inspect
import time
from collections.abc import Callable
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from loguru import Logger


def create_elapsed_logger[T, **P](logger: Logger) -> Callable[[Callable[P, T]], Callable[P, T]]:
    def decorator[T, **P](func: Callable[P, T]) -> Callable[P, T]:
        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            func_name = get_function_name(func, *args)
            start = time.perf_counter()
            result = func(*args, **kwargs)
            elapsed_time = f"{time.perf_counter() - start:.3f}"
            logger.info(
                f"{func_name} took {elapsed_time}", func_name=func_name, elasped_time=elapsed_time
            )
            return result

        @functools.wraps(func)
        async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            func_name = get_function_name(func, *args)
            start = time.perf_counter()
            result = await func(*args, **kwargs)
            elapsed_time = f"{time.perf_counter() - start:.3f}"
            logger.info(
                f"{func_name} took {elapsed_time}", func_name=func_name, elasped_time=elapsed_time
            )
            return result

        return async_wrapper if inspect.iscoroutinefunction(func) else wrapper

    return decorator


def log_elapsed[T](
    func_name: str,
    logger: Logger,
    func: Callable[[], T],
) -> T:
    start = time.perf_counter()
    result = func()
    elapsed_time = f"{time.perf_counter() - start:.3f}"
    logger.info(f"{func_name} took {elapsed_time}", func_name=func_name, elasped_time=elapsed_time)
    return result


def get_function_name[T, **P](func: Callable[P, T], *args) -> str:
    func_name = func.__name__
    return (
        f"{args[0].__class__.__name__}.{func_name}"
        if args and hasattr(args[0], "__class__") and hasattr(args[0], func_name)
        else func_name
    )



from typing import Any

from ..base import DictExpression, Operator, evaluate


class First(Operator):
    def __init__(self, expression: Any, /):
        self._expression = expression

    def expression(self) -> DictExpression:
        return {"$first": evaluate(self._expression)}


class MergeObjects(Operator):
    def __init__(self, *expressions: Any):
        self._expressions = list(expressions)

    def expression(self) -> DictExpression:
        expressions = [evaluate(e) for e in self._expressions]
        return {"$mergeObjects": expressions if len(expressions) > 1 else expressions[0]}


class AddToSet(Operator):
    def __init__(self, expression: Any, /):
        self._expression = expression

    def expression(self) -> DictExpression:
        return {"$addToSet": evaluate(self._expression)}


class Push(Operator):
    def __init__(self, expression: Any, /):
        self._expression = expression

    def expression(self) -> DictExpression:
        return {"$push": evaluate(self._expression)}



import math
from typing import Any, Self

from pydantic import Field

from . import model


class OffsetPaginationResult(model.BaseModel):
    items: list[Any]
    total: int
    page: int = Field(ge=1)
    per_page: int = Field(ge=1)
    extra: dict[str, Any] = Field(default_factory=dict)

    @property
    def prev_page(self) -> int | None:
        return self.page - 1 if self.has_prev else None

    @property
    def next_page(self) -> int | None:
        return self.page + 1 if self.has_next else None

    @property
    def has_prev(self) -> bool:
        return self.page > 1

    @property
    def has_next(self) -> bool:
        return self.page < self.pages

    @property
    def pages(self) -> int:
        return math.ceil(self.total / self.per_page)

    @classmethod
    def empty(cls, page: int, per_page: int) -> Self:
        return cls(total=0, items=[], page=page, per_page=per_page)

    @property
    def is_empty(self) -> bool:
        return not self.total


class CursorPaginationResult(model.BaseModel):
    items: list[Any]
    next_cursor: str | None = None

    @classmethod
    def empty(cls) -> Self:
        return cls(items=[])



from __future__ import annotations

import functools
import time
import types
from collections.abc import Callable
from typing import ClassVar, Self

import pendulum
import pymongo
from pymongo.collection import Collection
from pymongo.database import Database
from pymongo.errors import DuplicateKeyError

from ..types import Seconds
from .exceptions import LockAcquisitionFailedError

type MongoLockAcquirer = Callable[..., MongoLock]


class MongoLock:
    default_collection_name: ClassVar[str] = "compose.lock"
    index_name: ClassVar[str] = "expires_at_ttl"

    def __init__(
        self,
        key: str,
        collection: Collection,
        auto_release_after: Seconds = Seconds(60),
        timeout: Seconds = Seconds(60),
        lock_acquisition_interval: Seconds = Seconds(0.1),
        clock: Callable[[], pendulum.DateTime] = pendulum.DateTime.utcnow,
    ):
        self.key = key
        self.collection = collection
        self.auto_release_after = auto_release_after
        self.timeout = timeout
        self.lock_acquisition_interval = lock_acquisition_interval
        self.clock = clock

    @classmethod
    def _ensure_ttl_index(cls, collection: Collection) -> None:
        current_index = collection.index_information()
        if cls.index_name not in current_index:
            collection.create_index(
                [("expires_at", pymongo.ASCENDING)],
                name=cls.index_name,
                expireAfterSeconds=0,
            )

    @classmethod
    def acquirer(
        cls,
        db: Database,
        collection_name: str | None = None,
        clock: Callable[[], pendulum.DateTime] = pendulum.DateTime.utcnow,
    ) -> Callable[..., Self]:
        collection_name = collection_name or cls.default_collection_name
        collection = db[collection_name]
        cls._ensure_ttl_index(collection)
        return functools.partial(cls, collection=collection, clock=clock)

    def __enter__(self) -> Self:
        if self.acquire():
            return self

        raise LockAcquisitionFailedError(f"Failed to acquire lock for key: {self.key}")

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> None:
        self.release()

    def acquire(self) -> bool:
        try_until = self.clock().add(seconds=self.timeout)

        while True:
            now = self.clock()
            expires_at = now.add(seconds=self.auto_release_after)
            try:
                lock = self.collection.find_one_and_update(
                    {
                        "_id": self.key,
                        "expires_at": {"$lt": now},
                    },
                    {"$set": {"expires_at": expires_at}},
                    upsert=True,
                    return_document=pymongo.ReturnDocument.AFTER,
                )
                # MongoDB는 Date 타입에서 나노초를 버리기 때문에 <=로 비교
                if pendulum.instance(lock["expires_at"], tz=pendulum.UTC) <= expires_at:
                    return True

            except DuplicateKeyError:
                pass

            if self.clock() > try_until:
                return False

            time.sleep(self.lock_acquisition_interval)

    def release(self) -> None:
        self.collection.delete_one({"_id": self.key})



import collections
from collections.abc import Awaitable, Callable
from typing import Protocol

from compose.event import Event


class EventHandler(Protocol):
    async def handle(self, evt: Event) -> None: ...


class EventBus:
    def __init__(
        self, dependency_resolver: Callable[[type[EventHandler]], Awaitable[EventHandler]]
    ):
        self.dependency_resolver = dependency_resolver

        self._event_handlers: dict[str, set[type[EventHandler]]] = collections.defaultdict(set)

    async def handle_event(self, evt: Event) -> None:
        handler_types = self._event_handlers.get(evt.__class__.__name__, set())
        for handler_type in handler_types:
            handler = await self.dependency_resolver(handler_type)
            await handler.handle(evt)

    def register(
        self, event_cls: type[Event]
    ) -> Callable[[type[EventHandler]], type[EventHandler]]:
        def wrapper(handler_cls: type[EventHandler]) -> type[EventHandler]:
            event_name = event_cls.__name__

            registered_handler_types = self._event_handlers.get(event_name, set())
            if handler_cls in registered_handler_types:
                raise ValueError(
                    f"Handler `{handler_cls.__name__}` already registered for event `{event_name}`"
                )

            self._event_handlers[event_name].add(handler_cls)
            return handler_cls

        return wrapper



import uuid
from typing import Any, ClassVar, Generic

from pydantic import Field

from . import field, model
from .typing import IdT
from .utils import uuid7


class Entity(model.TimeStampedModel, Generic[IdT]):
    id: IdT

    updatable_fields: ClassVar[set[str]] = set()

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any) -> None:
        super().__pydantic_init_subclass__(**kwargs)

        fields = set(cls.model_fields.keys())
        if diff := set(cls.updatable_fields) - fields:
            raise ValueError(f"`updatable_fields` must be subset of {fields}, but got {diff}")

    def update(self, **kwargs: Any) -> None:
        for key, value in kwargs.items():
            if key not in self.updatable_fields:
                continue

            setattr(self, key, value)


try:
    from .types import PyObjectId

    class MongoEntity(Entity[PyObjectId]):
        id: PyObjectId = field.PyObjectIdField(default_factory=PyObjectId)

except ImportError:
    pass

try:
    from sqlalchemy.orm import Session  # noqa: F401

    class SQLEntity(Entity[uuid.UUID]):
        id: uuid.UUID = Field(default_factory=uuid7)

except ImportError:
    pass



import functools
from collections.abc import Callable
from typing import Any

from sqlalchemy.orm import Session, sessionmaker

from ..utils import unordered_partial


class SQLUnitOfWork[T]:
    def __init__(self, session_factory: sessionmaker[Session]):
        self.session_factory = session_factory

    def with_transaction(self, callback: Callable[..., T], **kwargs: Any) -> T:
        with self.session_factory() as session:
            try:
                result = unordered_partial(
                    p=functools.partial(callback, **kwargs),
                    t=Session,
                )(session)
                session.commit()
                return result
            except Exception:
                session.rollback()
                raise


def sql_transactional[T](func: Callable[..., T]) -> Callable[..., T]:
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> T:
        instance = args[0]

        if not hasattr(instance, "__dict__"):
            raise ValueError(f"`{instance.__class__.__name__}` does not have `__dict__` attribute")

        uow: SQLUnitOfWork | None = next(
            (t for t in instance.__dict__.values() if isinstance(t, SQLUnitOfWork)),
            None,
        )
        if uow is None:
            raise ValueError(
                f"`{instance.__class__.__name__}` does not have `{SQLUnitOfWork.__name__}` attribute"
            )

        return uow.with_transaction(functools.partial(func, *args, **kwargs))

    return wrapper



from __future__ import annotations

import urllib.parse
from typing import TYPE_CHECKING

from compose.types import ContentDisposition

if TYPE_CHECKING:
    import mypy_boto3_s3

type Seconds = int


class S3UrlGenerator:
    def __init__(
        self,
        s3_client: mypy_boto3_s3.Client,
        bucket: str | None = None,
        expires_in: Seconds = 3600,
    ):
        self.s3_client = s3_client
        self.bucket = bucket
        self.expires_in = expires_in

    def generate_download_url(
        self,
        key: str,
        filename: str | None = None,
        bucket: str | None = None,
        expires_in: Seconds | None = None,
        params: dict[str, str] | None = None,
    ) -> str:
        if (bucket := bucket or self.bucket) is None:
            raise ValueError("`bucket` must be provided if default bucket is not set")

        filename = urllib.parse.quote(filename or key.split("/")[-1])
        default_params = {
            "Bucket": bucket,
            "Key": key,
            "ResponseContentDisposition": ContentDisposition.attachment(filename),
        }
        return self.s3_client.generate_presigned_url(
            "get_object",
            Params=default_params | (params or {}),
            ExpiresIn=expires_in or self.expires_in,
        )

    def generate_upload_url(
        self,
        key: str,
        bucket: str | None = None,
        expires_in: Seconds | None = None,
        params: dict[str, str] | None = None,
    ) -> str:
        if (bucket := bucket or self.bucket) is None:
            raise ValueError("`bucket` must be provided if default bucket is not set")

        return self.s3_client.generate_presigned_url(
            "put_object",
            Params={"Bucket": bucket, "Key": key} | (params or {}),
            ExpiresIn=expires_in or self.expires_in,
        )



import abc
from typing import Any

from pydantic import Field

from compose.pagination import CursorPaginationResult, OffsetPaginationResult

from ..base import Query


class MongoQuery(Query, abc.ABC):
    @abc.abstractmethod
    def to_query(self) -> list[dict[str, Any]]:
        raise NotImplementedError


class MongoPaginationQuery[R](MongoQuery):
    @abc.abstractmethod
    def to_query(self) -> list[dict[str, Any]]:
        raise NotImplementedError

    @abc.abstractmethod
    def to_result(self, raw: dict[str, Any] | None) -> R:
        raise NotImplementedError


class MongoOffsetPaginationQuery(MongoPaginationQuery[OffsetPaginationResult]):
    page: int = Field(1, ge=1)
    per_page: int = Field(10, ge=1)

    @abc.abstractmethod
    def to_query(self) -> list[dict[str, Any]]:
        raise NotImplementedError

    def to_result(self, raw: dict[str, Any] | None) -> OffsetPaginationResult:
        if raw is None:
            return OffsetPaginationResult.empty(page=self.page, per_page=self.per_page)

        return OffsetPaginationResult(
            total=(raw["metadata"][0]["total"] if raw["metadata"] else 0),
            items=raw["items"],
            page=self.page,
            per_page=self.per_page,
        )


class MongoCursorPaginationQuery(MongoPaginationQuery[CursorPaginationResult], abc.ABC):
    per_page: int = 10

    @abc.abstractmethod
    def derive_cursor(self, last_item: dict[str, Any]) -> str:
        raise NotImplementedError

    def to_result(self, raw: dict[str, Any] | None) -> CursorPaginationResult:
        if raw is None:
            return CursorPaginationResult.empty()

        items: list[Any] = raw["items"]

        has_next = len(items) > self.per_page
        if has_next:
            items = items[: self.per_page]

        return CursorPaginationResult(
            items=items,
            next_cursor=self.derive_cursor(items[-1]) if has_next else None,
        )



from __future__ import annotations

import functools
from collections.abc import Callable
from typing import Any, Self

from .. import constants
from . import primitive


def _require_int(func: Callable[[type[Byte], Any], Byte]) -> Callable[[type[Byte], Any], Byte]:
    @functools.wraps(func)
    def wrapper(cls, v: Any, /) -> Byte:
        if not isinstance(v, int):
            raise TypeError(f"`{cls.__name__}.{func.__name__}` expects int, got {type(v).__name__}")
        return func(cls, v)

    return wrapper


class Byte(primitive.Int):
    def __new__(cls, v: Any, /) -> Self:
        v = super().__new__(cls, v)
        if v < 0:
            raise ValueError(f"`{cls.__name__}` must be a non-negative integer")
        return v

    @classmethod
    @_require_int
    def from_mib(cls, v: int, /) -> Self:
        return cls(v * constants.Unit.MIB)

    @classmethod
    @_require_int
    def from_gib(cls, v: int, /) -> Self:
        return cls(v * constants.Unit.GIB)



import concurrent.futures
import functools
from collections.abc import Callable, Hashable


def execute_in_pool[K, T](
    pool_factory: Callable[[], concurrent.futures.Executor],
    funcs: dict[K, functools.partial[T]],
    timeout: int | None = None,
) -> dict[K, T]:
    result = {}
    with pool_factory() as executor:
        future_to_key = dict()
        for key, func in funcs.items():
            future = executor.submit(func)
            future_to_key[future] = key

        for future in concurrent.futures.as_completed(future_to_key, timeout=timeout):
            result[future_to_key[future]] = future.result()

    return result


class ThreadPoolJob[K: Hashable, **P, T]:
    def __init__(
        self,
        key: K,
        func: Callable[P, T],
        *args: P.args,
        **kwargs: P.kwargs,
    ):
        self.key = key
        self.func = func
        self.args = args
        self.kwargs = kwargs


class ThreadPoolExecutor:
    def __init__(self, max_workers: int, timeout: int | None = None):
        self.max_workers = max_workers
        self.timeout = timeout

    def execute[K, **P, T](
        self,
        jobs: list[ThreadPoolJob[K, P, T]],
        max_workers: int | None = None,
        timeout: int | None = None,
    ) -> dict[K, T]:
        result: dict[K, T] = {}
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=max_workers or self.max_workers
        ) as executor:
            future_to_key: dict[concurrent.futures.Future[T], K] = {}

            for job in jobs:
                future = executor.submit(functools.partial(job.func, *job.args, **job.kwargs))
                future_to_key[future] = job.key

            for future in concurrent.futures.as_completed(
                future_to_key, timeout=timeout or self.timeout
            ):
                result[future_to_key[future]] = future.result()

        return result



from .base import BaseRepository

__all__ = ["BaseRepository"]

try:
    from .mongo import MongoDocument, MongoRepository, setup_indexes
    from .shortcut import finder, lister

    __all__ += ["MongoDocument", "MongoRepository", "finder", "lister", "setup_indexes"]
except ImportError:
    pass

try:
    from .sql import SQLRepository

    __all__ += ["SQLRepository"]
except ImportError:
    pass



import inspect
import types
from collections.abc import Callable
from typing import Annotated, Any, Union, get_args, get_origin

from dishka import AsyncContainer, FromDishka
from dishka.integrations.fastapi import DishkaRoute
from dishka.registry import Registry
from fastapi.routing import APIRoute


def injected_route(container: AsyncContainer) -> type[APIRoute]:
    resolvable = _collect_resolvable_types(container)

    class InjectedDishkaRoute(DishkaRoute):
        def __init__(self, path: str, endpoint: Callable[..., Any], **kwargs: Any):
            endpoint = _convert_signature(endpoint, resolvable)
            super().__init__(path, endpoint, **kwargs)

    return InjectedDishkaRoute


def _collect_resolvable_types(container: AsyncContainer) -> set[type]:
    result: set[type] = set()
    registry: Registry | None = container.registry
    while registry is not None:
        for key in registry.factories:
            if isinstance(key.type_hint, type):
                result.add(key.type_hint)
        registry = getattr(registry, "child_registry", None)
    result.discard(AsyncContainer)
    return result


def _is_excluded_type(hint: Any) -> bool:
    if hint is inspect.Parameter.empty or hint is Any:
        return True
    origin = get_origin(hint)
    if origin is Annotated:
        return True
    if origin is Union or origin is types.UnionType:
        return True
    return False


def _convert_signature(
    endpoint: Callable[..., Any],
    resolvable: set[type],
) -> Callable[..., Any]:
    sig = inspect.signature(endpoint)
    hints = {k: v for k, v in inspect.get_annotations(endpoint).items() if k != "return"}

    new_params = []
    new_annotations = {}
    changed = False

    for name, param in sig.parameters.items():
        hint = hints.get(name, inspect.Parameter.empty)

        if _is_excluded_type(hint):
            new_params.append(param)
            if name in hints:
                new_annotations[name] = hints[name]
            continue

        bare_type = get_args(hint)[0] if get_origin(hint) else hint
        if bare_type in resolvable:
            from_dishka = FromDishka[bare_type]
            new_params.append(
                param.replace(annotation=from_dishka, default=inspect.Parameter.empty)
            )
            new_annotations[name] = from_dishka
            changed = True
        else:
            new_params.append(param)
            if name in hints:
                new_annotations[name] = hints[name]

    if not changed:
        return endpoint

    if "return" in inspect.get_annotations(endpoint):
        new_annotations["return"] = inspect.get_annotations(endpoint)["return"]

    endpoint.__signature__ = sig.replace(parameters=new_params)  # type: ignore[attr-defined]
    endpoint.__annotations__ = new_annotations  # type: ignore[attr-defined]
    return endpoint



from collections.abc import Generator
from typing import Self

import httpx


class HeaderAuth(httpx.Auth):
    def __init__(self, secrets: dict[str, str]) -> None:
        self.secrets = secrets

    def auth_flow(self, request: httpx.Request) -> Generator[httpx.Request, httpx.Response, None]:
        for key, value in self.secrets.items():
            request.headers[key] = value
        yield request

    @classmethod
    def single(cls, key: str, header_name: str = "x-api-key") -> Self:
        return cls({header_name: key})



from typing import Any

from ..base import Operator, evaluate
from ..types import DictExpression


class RegexMatch(Operator):
    def __init__(self, field: Any, value: Any):
        self.field = field
        self.value = value

    def expression(self) -> DictExpression:
        return {"$regexMatch": {"input": evaluate(self.field), "regex": evaluate(self.value)}}



import inspect
from collections.abc import Callable
from dataclasses import asdict
from typing import Annotated, Any, cast, get_args

import pydantic_core
from fastapi import Depends, Path, Query, params
from pydantic import BaseModel, Field, Json

from compose import query, types


def dict_to_json(v: dict[str, Any] | None) -> str | None:
    if v is None:
        return v

    return pydantic_core.to_json(v).decode()


TYPE_VALIDATORS = {
    Json: [
        dict_to_json,
    ]
}


_QUERY_FIELD_ARGS = ("title", "alias", "default", "default_factory", "description")


def _build_query_resolver[Q: BaseModel](q: type[Q], /) -> Callable[..., Q]:
    pre_validators: dict[str, list[Callable[..., Any]]] = {}
    for field_name, field_info in q.model_fields.items():  # type: ignore[arg-type]
        for arg in get_args(field_info.annotation):
            for validator in TYPE_VALIDATORS.get(arg, []):
                pre_validators.setdefault(field_name, []).append(validator)

    def factory(**kwargs: Any) -> Q:
        for key, validators in pre_validators.items():
            if key in kwargs:
                for validate in validators:
                    kwargs[key] = validate(kwargs[key])
        return q(**kwargs)

    cast(Any, factory).__signature__ = inspect.Signature(
        parameters=[
            inspect.Parameter(
                field_name,
                inspect.Parameter.KEYWORD_ONLY,
                default=Query(
                    **{arg: getattr(field_info, arg, None) for arg in _QUERY_FIELD_ARGS},  # type: ignore[arg-type]
                    **{k: v for item in field_info.metadata for k, v in asdict(item).items()},
                ),
                annotation=field_info.annotation,
            )
            for field_name, field_info in q.model_fields.items()  # type: ignore[arg-type]
        ],
    )
    return factory


def as_query[Q: BaseModel](q: type[Q], /) -> Any:
    return Depends(_build_query_resolver(q))


def create_model_dependency_resolver[T: BaseModel](
    model_type: type[T],
    dependencies: dict[str, tuple[type, Any]],
) -> Callable[..., Any]:
    dep_names = list(dependencies.keys())

    def wrapper(t, **kwargs) -> T:
        return t.model_copy(update={name: kwargs[name] for name in dep_names}, deep=True)

    cast(Any, wrapper).__signature__ = inspect.Signature(
        parameters=[
            inspect.Parameter("t", inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=model_type),
            *(
                inspect.Parameter(
                    name, inspect.Parameter.KEYWORD_ONLY, default=default, annotation=type_
                )
                for name, (type_, default) in dependencies.items()
            ),
        ],
    )
    return wrapper


def with_fields[T: BaseModel](model_type: type[T], **kwargs: Any) -> Any:
    return Depends(create_model_dependency_resolver(model_type, kwargs))


class FromPath:
    @classmethod
    def object_id(
        cls, path: params.Path | None = None
    ) -> "tuple[type[types.PyObjectId], params.Path]":
        if not hasattr(types, "PyObjectId"):
            raise ImportError("FromPath.object_id requires `bson`. Install `pymongo`")
        return types.PyObjectId, path or Path(...)

    @classmethod
    def int(cls, path: params.Path | None = None) -> tuple[type[int], params.Path]:
        return int, path or Path(...)

    @classmethod
    def str(cls, path: params.Path | None = None) -> tuple[type[str], params.Path]:
        return str, path or Path(...)


class FromAuth[U]:
    def __init__(self, user_getter: Callable[..., U]) -> None:
        self._user_getter = user_getter

    def field(self, attr: str, field_type: type) -> tuple[type, params.Depends]:
        user_getter = self._user_getter

        def get_field(user: U = Depends(user_getter)):
            return getattr(user, attr)

        return field_type, Depends(get_field)


class _OffsetPaginationParams(query.Query):
    page: int = Field(1, ge=1)
    per_page: int = Field(10, ge=1)


OffsetPaginationParams = Annotated[_OffsetPaginationParams, as_query(_OffsetPaginationParams)]



from collections.abc import Awaitable, Callable
from typing import Any

from dishka import AsyncContainer


def create_event_handler_resolver(container: AsyncContainer) -> Callable[[type], Awaitable[Any]]:
    async def resolver(handler_type: type) -> Any:
        return await container.get(handler_type)

    return resolver



__all__: list[str] = []

try:
    from .mongo import MongoUnitOfWork, mongo_transactional

    __all__ += ["MongoUnitOfWork", "mongo_transactional"]
except ImportError:
    pass

try:
    from .sql import SQLUnitOfWork, sql_transactional

    __all__ += ["SQLUnitOfWork", "sql_transactional"]
except ImportError:
    pass
